from functools import partial, wraps
from typing import Callable, Hashable, TypeVar

from equinox import filter_pmap, filter_vmap

from .base import Functional


T = TypeVar("T")


def pipe_vmap(
    func: Callable[..., T] = None,
    *,
    in_axes=0,
    out_axes=0,
    axis_name: Hashable = None,
    axis_size: int = None,
    name: str = None
) -> Callable[..., T]:
    """Creates a functional from a function with vmap.

    Args:
        func: The function to be wrapped.
        in_axes: The number of input axes.
        out_axes: The number of output axes.
        axis_name: The name of the axis.
        axis_size: The size of the axis.
        name: The name of the functional. If not provided, the name of the
        function is used.

    Returns:
        A callable that creates a functional from the given function.

    Example:
        >>> from stamox import pipe_vmap
        >>> f = lambda x: x + 1
        >>> f = pipe_vmap(f)
        >>> g = f >> f >> f
        >>> g(jnp.array([1, 2, 3]))
        Array([4, 5, 6], dtype=int32)
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        if isinstance(func, Functional):
            if func.func is not None:
                func = func.func
        fn = filter_vmap(
            func,
            in_axes=in_axes,
            out_axes=out_axes,
            axis_name=axis_name,
            axis_size=axis_size,
        )
        return Functional(name=name, fn=fn)

    return wrap if func is None else wrap(func)


def partial_pipe_vmap(
    func: Callable[..., T] = None, *, name: str = None
) -> Callable[..., T]:
    """Partially apply a function to a vmap.

    Args:
        func (Callable[P, T]): The function to partially apply.
        name (str, optional): The name of the function. Defaults to None.

    Returns:
        Callable[P, T]: A partially applied function.

    Example:
        >>> from stamox import partial_pipe_vmap
        >>> f = lambda x, y: x + y
        >>> f = partial_pipe_vmap(f)
        >>> g = f(y=1) >> f(y=2) >> f(y=3)
        >>> g(jnp.array([1, 2, 3]))
        Array([7, 8, 9], dtype=int32)
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        else:
            name = func.__name__

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        if isinstance(func, Functional):
            if func.func is not None:
                if func.pipe_type == "pmap" or func.pipe_type == "vmap":
                    raise ValueError(
                        "You can not use pipe_pmap or pipe_vmap with partial_pipe_*, use make_pipe or pipe_* instead."
                    )
                func = func.func

        @wraps(func)
        def partial_fn(
            *args,
            in_axes=0,
            out_axes=0,
            axis_name: Hashable = None,
            axis_size: int = None,
            **kwargs
        ):
            fn = partial(func, **kwargs)
            fn = filter_vmap(
                fn,
                in_axes=in_axes,
                out_axes=out_axes,
                axis_name=axis_name,
                axis_size=axis_size,
            )
            if len(args) != 0:
                return fn(*args)
            return Functional(name=name, fn=fn, pipe_type="vmap")

        return partial_fn

    return wrap if func is None else wrap(func)


def pipe_pmap(
    func: Callable[..., T] = None,
    *,
    in_axes=0,
    out_axes=0,
    axis_name: Hashable = None,
    axis_size: int = None,
    name: str = None
) -> Callable[..., T]:
    """Creates a functional object from a given function.

    Args:
        func (Callable[P, T]): The function to be wrapped.
        in_axes (int): The number of input axes for the function.
        out_axes (int): The number of output axes for the function.
        axis_name (Hashable): The name of the axis.
        axis_size (int ): The size of the axis.
        name (str): The name of the functional object.

    Returns:
        Callable[P, T]: A callable object that wraps the given function.

    Example:
        >>> from stamox import pipe_pmap
        >>> f = lambda x: x + 1
        >>> f = pipe_pmap(f)
        >>> g = f >> f >> f
        >>> g(jnp.array([1, 2, 3]))
        Array([4, 5, 6], dtype=int32)
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        if isinstance(func, Functional):
            if func.func is not None:
                if func.pipe_type == "pmap" or func.pipe_type == "vmap":
                    raise ValueError(
                        "You can not use pipe_pmap or pipe_vmap with partial_pipe_*, use make_pipe or pipe_* instead."
                    )
                func = func.func
        fn = filter_pmap(
            func,
            in_axes=in_axes,
            out_axes=out_axes,
            axis_name=axis_name,
            axis_size=axis_size,
        )

        return Functional(name=name, fn=fn, pipe_type="pmap")

    return wrap if func is None else wrap(func)


def partial_pipe_pmap(
    func: Callable[..., T] = None, *, name: str = None
) -> Callable[..., T]:
    """Partially apply a function to a pipe.

    Args:
        func (Callable[P, T]): The function to partially apply.
        name (str, optional): The name of the function. Defaults to None.

    Returns:
        Callable[P, T]: A partially applied function.

    Example:
        >>> from stamox import partial_pipe_pmap
        >>> f = lambda x, y: x + y
        >>> f = partial_pipe_pmap(f)
        >>> g = f(y=1) >> f(y=2) >> f(y=3)
        >>> g(jnp.array([1, 2, 3]))
        Array([7, 8, 9], dtype=int32)
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        if isinstance(func, Functional):
            if func.func is not None:
                func = func.func

        @wraps(func)
        def partial_fn(
            *args,
            in_axes=0,
            out_axes=0,
            axis_name: Hashable = None,
            axis_size: int = None,
            **kwargs
        ):
            fn = partial(func, **kwargs)
            fn = filter_pmap(
                fn,
                in_axes=in_axes,
                out_axes=out_axes,
                axis_name=axis_name,
                axis_size=axis_size,
            )
            if len(args) != 0:
                return fn(*args, **kwargs)
            return Functional(name=name, fn=fn, pipe_type="pmap")

        return partial_fn

    return wrap if func is None else wrap(func)



from functools import partial, wraps
from typing import Callable, TypeVar

from equinox import filter_jit

from .base import Functional


T = TypeVar("T")


def pipe_jit(
    func: Callable[..., T] = None, *, donate: str = "none", name: str = None
) -> Callable[..., T]:
    """Creates a pipeable jitted functional from a given function.

    Args:
        func: The function to create the functional from.
        donate: Optional donation string.
        name: Optional name for the functional.

    Returns:
        A callable that creates a functional from the given function.

    Example:
        >>> from stamox import pipe_jit
        >>> f = lambda x: x + 1
        >>> f = pipe_jit(f)
        >>> g = f >> f >> f
        >>> g(1)
        4
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        fn = filter_jit(func, donate=donate)

        return Functional(name=name, fn=fn)

    return wrap if func is None else wrap(func)


def partial_pipe_jit(
    func: Callable[..., T] = None, *, name: str = None
) -> Callable[..., T]:
    """Creates a partial pipeable jitted functional from a given function.

    Args:
        func (Callable[P, T]): _description_
        name (str, optional): _description_. Defaults to None.

    Returns:
        a partial pipeable jitted functional from a given function.

    Example:
        >>> from stamox.core import partial_pipe_jit
        >>> f = lambda x, y: x + y
        >>> f = partial_pipe_jit(f)
        >>> g = f(y=1) >> f(y=2) >> f(y=3)
        >>> g(1)
        7
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        if isinstance(func, Functional):
            if func.func is not None:
                if func.pipe_type == "pmap" or func.pipe_type == "vmap":
                    raise ValueError(
                        "You can not use pmap or vmap with partial_pipe_*, use make_pipe or pipe_* instead."
                    )
                func = func.func

        @wraps(func)
        def partial_fn(*args, donate: str = "none", **kwargs):
            fn = filter_jit(func, donate=donate)
            fn = partial(fn, **kwargs)
            if len(args) != 0:
                return fn(*args, **kwargs)
            return Functional(name=name, fn=fn, pipe_type="jit")

        return partial_fn

    return wrap if func is None else wrap(func)



from jaxtyping import ArrayLike

from ..core import StateFunc


class HypoTest(StateFunc):
    _statistic: ArrayLike
    _parameters: ArrayLike
    _p_value: ArrayLike
    _estimate: ArrayLike
    _null_value: ArrayLike
    _alternative: str

    def __init__(
        self,
        statistic=None,
        parameters=None,
        p_value=None,
        estimate=None,
        null_value=None,
        alternative=None,
        name="HTest",
    ):
        super().__init__(name=name, fn=None)
        self._statistic = statistic
        self._parameters = parameters
        self._p_value = p_value
        self._estimate = estimate
        self._null_value = null_value
        self._alternative = alternative

    @property
    def statistic(self):
        return self._statistic

    @property
    def parameters(self):
        return self._parameters

    @property
    def p_value(self):
        return self._p_value



"""Stamox: Stats Models in JAX Library"""

import sys

from stamox import (
    anova,
    basic,
    cluster,
    core,
    correlation,
    distribution,
    experimental,
    formula,
    functions,
    hypothesis,
    math,
    pipe_functions,
    regression,
    sample,
    transformation,
)
from stamox.core import (
    better_partial,
    Functional,
    make_partial_pipe,
    make_pipe,
    partial_pipe_jit,
    partial_pipe_pmap,
    partial_pipe_vmap,
    Pipe,
    pipe_jit,
    pipe_pmap,
    pipe_vmap,
    Pipeable,
    StateFunc,
)


def _check_py_version():
    if sys.version_info[0] < 3:
        raise Exception("Please use Python 3. Python 2 is not supported.")


_check_py_version()


__all__ = [
    "anova",
    "basic",
    "cluster",
    "core",
    "correlation",
    "distribution",
    "experimental",
    "formula",
    "functions",
    "hypothesis",
    "math",
    "pipe_functions",
    "regression",
    "sample",
    "transformation",
    "Functional",
    "make_partial_pipe",
    "make_pipe",
    "partial_pipe_jit",
    "partial_pipe_pmap",
    "partial_pipe_vmap",
    "pipe_jit",
    "pipe_pmap",
    "pipe_vmap",
    "Pipe",
    "Pipeable",
    "StateFunc",
    "better_partial",
]



import jax.numpy as jnp
from jax import vmap
from jaxtyping import ArrayLike


mean = jnp.mean
sd = jnp.std
var = jnp.var
median = jnp.median
quantile = jnp.quantile
min = jnp.min
max = jnp.max
sum = jnp.sum
prod = jnp.prod
cumsum = jnp.cumsum
cumprod = jnp.cumprod
diff = jnp.diff


def scale(x: ArrayLike, axis: int = 0, dtype=jnp.float_) -> ArrayLike:
    """Calculate standardized x along axis.

    Args:
        x (array-like): Input array.
        axis (int, optional): Axis along which to calculate mean and standard deviation. Defaults to 0.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: Standardized x along axis.
    """
    x = jnp.asarray(x, dtype=dtype)
    _mean = mean(x, axis=axis)
    _std = sd(x, axis=axis, ddof=1)
    _scaled = vmap(lambda a, b, c: (a - b) / c, in_axes=(axis, None, None))(
        x, _mean, _std
    )
    return _scaled



from ._func_state import predict, summary
from .base import (
    Functional,
    StateFunc,
)
from .better_partial import better_partial
from .jit import partial_pipe_jit, pipe_jit
from .maps import partial_pipe_pmap, partial_pipe_vmap, pipe_pmap, pipe_vmap
from .pipe import (
    make_partial_pipe,
    make_pipe,
    Pipe,
    Pipeable,
)


__all__ = [
    "better_partial",
    "Functional",
    "make_partial_pipe",
    "make_pipe",
    "partial_pipe_jit",
    "partial_pipe_pmap",
    "partial_pipe_vmap",
    "Pipe",
    "Pipeable",
    "pipe_jit",
    "pipe_pmap",
    "pipe_vmap",
    "predict",
    "StateFunc",
    "summary",
]



from .anova import one_way
from .basic import (
    cumprod,
    cumsum,
    diff,
    max,
    mean,
    median,
    min,
    prod,
    quantile,
    scale,
    sd,
    sum,
    var,
)
from .cluster import kmeans
from .core import (
    predict,
    summary,
)
from .correlation import cor, pearsonr, spearmanr
from .decomposition import princomp
from .distribution import (
    dbeta,
    dbinom,
    dcauchy,
    dchisq,
    dexp,
    dF,
    dgamma,
    dlaplace,
    dnorm,
    dpareto,
    dpoisson,
    dt,
    dunif,
    dweibull,
    ecdf,
    pbeta,
    pbinom,
    pcauchy,
    pchisq,
    pexp,
    pF,
    pgamma,
    plaplace,
    pnorm,
    ppareto,
    ppoisson,
    pt,
    punif,
    pweibull,
    qbeta,
    qbinom,
    qcauchy,
    qchisq,
    qexp,
    qF,
    qgamma,
    qlaplace,
    qnorm,
    qpareto,
    qpoisson,
    qt,
    qunif,
    qweibull,
    rbeta,
    rbinom,
    rcauchy,
    rchisq,
    rexp,
    rF,
    rgamma,
    rlaplace,
    rnorm,
    rpareto,
    rpoisson,
    rt,
    runif,
    rweibull,
)
from .hypothesis import (
    bartlett_test,
    durbin_watson_test,
    friedman_test,
)
from .regression import lm
from .sample import bootstrap, bootstrap_sample, jackknife, jackknife_sample
from .transformation import boxcox, z_fisher


__all__ = [
    "one_way",
    "kmeans",
    "spearmanr",
    "pearsonr",
    "cor",
    "bartlett_test",
    "durbin_watson_test",
    "friedman_test",
    "lm",
    "bootstrap",
    "bootstrap_sample",
    "jackknife",
    "jackknife_sample",
    "boxcox",
    "z_fisher",
    "dbeta",
    "dbinom",
    "dcauchy",
    "dchisq",
    "dexp",
    "dF",
    "dgamma",
    "dlaplace",
    "dnorm",
    "dpareto",
    "dpoisson",
    "dt",
    "dunif",
    "dweibull",
    "ecdf",
    "pbeta",
    "pbinom",
    "pcauchy",
    "pchisq",
    "pexp",
    "pF",
    "pgamma",
    "plaplace",
    "pnorm",
    "ppareto",
    "ppoisson",
    "pt",
    "punif",
    "pweibull",
    "qbeta",
    "qbinom",
    "qcauchy",
    "qchisq",
    "qexp",
    "qF",
    "qgamma",
    "qlaplace",
    "qnorm",
    "qpareto",
    "qpoisson",
    "qt",
    "qunif",
    "qweibull",
    "rbeta",
    "rbinom",
    "rcauchy",
    "rchisq",
    "rexp",
    "rF",
    "rgamma",
    "rlaplace",
    "rnorm",
    "rpareto",
    "rpoisson",
    "rt",
    "runif",
    "rweibull",
    "predict",
    "summary",
    "mean",
    "sd",
    "var",
    "median",
    "quantile",
    "min",
    "max",
    "sum",
    "prod",
    "cumsum",
    "cumprod",
    "diff",
    "scale",
    "princomp",
]



import inspect
from typing import TypeVar


_T = TypeVar("_T")

Self = TypeVar("Self", bound="better_partial")


class better_partial:
    __slots__ = "func", "keywords", "__dict__", "__weakref__"

    def __new__(cls, func, /, *args, **keywords) -> Self:
        if not callable(func):
            raise TypeError("the first argument must be callable")

        if hasattr(func, "func"):
            keywords = {**func.keywords, **keywords}
            func = func.func
        self = super(better_partial, cls).__new__(cls)
        # get all args name
        args_name = inspect.getfullargspec(func).args
        keywords_default = {}
        self.args_name = args_name
        self.func = func
        unassign_argsname = [arg for arg in args_name if arg not in keywords]
        # get default argnames and value
        sig = inspect.signature(func)
        for param in sig.parameters.values():
            if param.default != inspect.Parameter.empty:
                keywords_default[param.name] = param.default
        for i, arg in enumerate(unassign_argsname):
            if i < len(args):
                keywords[arg] = args[i]
        self.keywords = keywords_default
        self.keywords.update(keywords)

        return self

    def __call__(self, *args, **keywords) -> _T:
        keywords = {**self.keywords, **keywords}
        unassign_or_none_argsname = [
            arg
            for arg in self.args_name
            if arg not in keywords or keywords[arg] is None
        ]
        for i, arg in enumerate(unassign_or_none_argsname):
            if i < len(args):
                keywords[arg] = args[i]
        return self.func(**keywords)



from typing import List, Tuple, Union

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from jax import vmap
from jaxtyping import ArrayLike
from pandas import DataFrame

from ..distribution import pF, pt
from ..formula import get_design_matrices
from ._base import RegState


class OLSState(RegState):
    _X_names: List[str]
    _y_names: List[str]
    _SSE: float
    _SSR: float
    _SST: float
    _coef_std: ArrayLike

    def __init__(
        self,
        in_features,
        out_features,
        coefs,
        df_resid,
        df_model,
        n_obs,
        resid,
        fitted_values,
        rank,
        SSE,
        SSR,
        SST,
        coefs_std,
        X_names=None,
        y_names=None,
        dtype=jnp.float32,
        name="OLSResult",
    ):
        super().__init__(
            in_features,
            out_features,
            coefs,
            df_resid,
            df_model,
            n_obs,
            resid,
            fitted_values,
            rank,
            dtype=dtype,
            name=name,
        )
        if X_names is None:
            X_names = ["Intercept"] + [f"X{i}" for i in range(in_features)]
        self._X_names = X_names
        if y_names is None:
            y_names = [f"y{i}" for i in range(out_features)]
        self._y_names = y_names
        self._SSE = SSE
        self._SSR = SSR
        self._SST = SST
        self._coef_std = coefs_std

    @property
    def X_names(self):
        return self._X_names

    @property
    def y_names(self):
        return self._y_names

    @property
    def SSE(self):
        return self._SSE

    @property
    def SSR(self):
        return self._SSR

    @property
    def SST(self):
        return self._SST

    @property
    def f_value(self):
        return (self.SSR / self._df_model) / (self.SSE / self._df_resid)

    @property
    def f_pvalue(self):
        return 1 - pF(self.f_value, self._df_model, self._df_resid, dtype=self.dtype)

    @property
    def r2(self):
        return self.SSR / self.SST

    @property
    def r2_adj(self):
        return 1 - (self.SSE / self._df_resid) / (self.SST / (self._n_obs - 1))

    @property
    def coef_std(self):
        return self._coef_std

    @property
    def std_err(self):
        return jnp.dot(self.coef_std, self.resid_stderr).ravel()

    @property
    def t_values(self):
        return jnp.divide(self.coefs.ravel(), self.std_err.ravel())

    @property
    def p_values(self):
        return (
            pt(
                jnp.abs(self.t_values),
                self.df_resid,
                lower_tail=False,
                dtype=self.dtype,
            )
            * 2
        )

    @property
    def resid_stderr(self):
        return jnp.sqrt(self.SSE / self.df_resid)

    @property
    def confint(self):
        rec = jnp.array([-1, 1]) * pt(0.975, self.df_resid, dtype=self.dtype)
        return vmap(lambda c, ci, z: c + ci * z, in_axes=(0, 0, None))(
            self.coefs,
            self.std_err,
            rec,
        )

    def _summary(self):
        summary_header = """
        Linear Regression Model Summary
        --------------------------------------------

        R-squared:          {r_sq:.4f}
        Adjusted R-squared: {adj_r_sq:.4f}
        F-statistic:        {f_value:.4f} on {df_model} and {df_resid} DF, p-value: {f_pvalue:.4f}
        SSE :               {SSE:.4f}
        SSR :               {SSR:.4f}
        Coefficients:
        {:<14} {:>10} {:>10} {:>10} {:>10} {:>10} {:>10}
        """.format(
            "",
            "Estimate",
            "Std.Error",
            "t-value",
            "p-value",
            "0.025",
            "0.975",
            r_sq=self.r2,
            adj_r_sq=self.r2_adj,
            f_value=self.f_value,
            df_model=self.df_model,
            df_resid=self.df_resid,
            f_pvalue=self.f_pvalue,
            SSE=self.SSE,
            SSR=self.SSR,
        )  # noqa: E501
        print(summary_header)
        new_line = "        " + "-" * 80
        print(new_line)
        for i, name in enumerate(self.X_names):
            p_value_formatted = (
                f"{self.p_values[i]:.4f}"
                if self.p_values[i] >= 0.001
                else f"{self.p_values[i]:.1e}"
            )
            summary_row = "        {:<14} {:>10.4f} {:>10.4f} {:>10.4f} {:>10} {:>10.4f} {:>10.4f}".format(
                name,
                self.coefs[i],
                self.std_err[i],
                self.t_values[i],
                p_value_formatted,
                self.confint[i, 0],
                self.confint[i, 1],
            )
            print(summary_row)


def lm(
    data: Union[List, Tuple, DataFrame, ArrayLike],
    formula=None,
    subset=None,
    weights=None,
    NA_action="drop",
    method="qr",
    dtype: jnp.dtype = jnp.float_,
) -> OLSState:
    """Fits a linear model using the given data and parameters.

    Args:
        data (Union[List, Tuple, DataFrame, ArrayLike]): The data to fit the linear model with.
        formula (str, optional): A formula for the linear model. Defaults to None.
        subset (list, optional): A list of indices to use as a subset of the data. Defaults to None.
        weights (array-like, optional): An array of weights to apply to the data. Defaults to None.
        NA_action (str, optional): The action to take when encountering missing values. Defaults to "drop".
        method (str, optional): The method to use for fitting the linear model. Defaults to "qr".
        dtype (jnp.float——, optional): The data type to use for the linear model. Defaults to jnp.float_.

    Returns:
        OLSState: The state of the fitted linear model.

    Example:
        >>> from stamox.functions import lm
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(42)
        >>> X = np.random.uniform(size=(1000, 3))
        >>> y = (
                3 * X[:, 0]
                + 2 * X[:, 1]
                - 7 * X[:, 2]
                + 1.
            )
        >>> data = pd.DataFrame(
                np.concatenate([X, y.reshape((-1, 1))], axis=1),
                columns=["x1", "x2", "x3", "y"],
            )
        >>> res = lm(data, "y ~ x1 + x2 + x3")
        >>> res.coefs
        Array([ 1., 3.,  2., -7.], dtype=float32)

    """
    if formula is None:
        if isinstance(data, DataFrame):
            raise ValueError("formula is required when data is DataFrame")
        elif isinstance(data, list):
            y = data[0]
            X = data[1]
        elif isinstance(data, tuple):
            y = data[0]
            X = data[1]
        elif isinstance(data, jnp.ndarray) or isinstance(data, np.ndarray):
            y = data[:, 0]
            X = data[:, 1:]
        else:
            raise ValueError("data is not supported")

        if dtype is None:
            dtype = jnp.promote_types(y.dtype, X.dtype)
        y = jnp.asarray(y, dtype=dtype)
        X = jnp.asarray(X, dtype=dtype)
        X_names = None
        y_names = None
    else:
        matrices = get_design_matrices(data, formula, NA_action=NA_action, dtype=dtype)
        y = matrices.y
        X = matrices.X
        X_names = matrices.X_names
        y_names = matrices.y_names
    if subset is not None:
        y = y[subset, :]
        X = X[subset, :]
    if weights is not None:
        weights = jnp.asarray(weights, dtype=dtype).reshape(-1, 1)
        W = jnp.diag(weights.reshape(-1))
        y = jnp.multiply(y, weights)
        X = jnp.matmul(X.T, W).T
    return _fit_lm(X, y, method, X_names=X_names, y_names=y_names)


@eqx.filter_jit
def _fit_lm(X, y, method, X_names=None, y_names=None):
    num_samples = X.shape[0]
    in_features = X.shape[1]
    out_features = 1
    if len(jnp.shape(y)) == 2:
        if jnp.shape(y)[1] > 1:
            raise NotImplementedError("multiple output is not supported yet.")
    elif len(jnp.shape(y)) == 1:
        y = jnp.atleast_1d(y)
    else:
        raise NotImplementedError("batchable output is not supported yet.")
    if method == "inv":
        X_pinv = jnp.linalg.inv(X.T @ X) @ X.T
    elif method == "qr":
        Q, R = jnp.linalg.qr(X)
        R_inv = jnp.linalg.inv(R)
        X_pinv = R_inv @ Q.T
    elif method == "svd":
        U, S, Vt = jnp.linalg.svd(X)
        D = jnp.zeros((num_samples, in_features), dtype=X.dtype)
        D = D.at[:in_features, :in_features].set(jnp.diag(1 / S))
        X_pinv = Vt.T @ D.T @ U.T
    else:
        raise NotImplementedError(f"Not Implemented method {method}")
    coefs = X_pinv @ y
    _coefs_std = jnp.sqrt(jnp.diag(X_pinv @ X_pinv.T)).reshape(-1, 1)
    _fitted_values = X @ coefs
    _residuals = y - _fitted_values
    _df_resid = num_samples - in_features
    _df_model = in_features - 1
    _rank = jnp.linalg.matrix_rank(X)
    _SSE = jnp.sum(_residuals**2)
    _SST = jnp.sum((y - jnp.mean(y)) ** 2)
    _SSR = _SST - _SSE
    state = OLSState(
        in_features=in_features,
        out_features=out_features,
        coefs=coefs,
        df_resid=_df_resid,
        df_model=_df_model,
        n_obs=num_samples,
        resid=_residuals,
        fitted_values=_fitted_values,
        rank=_rank,
        SSE=_SSE,
        SSR=_SSR,
        SST=_SST,
        coefs_std=_coefs_std,
        dtype=X.dtype,
        X_names=X_names,
        y_names=y_names,
    )
    return state



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_jit
from jax import pure_callback, ShapeDtypeStruct
from jax._src.random import Shape
from jax.random import KeyArray
from jax.scipy.special import gammainc, gammaln
from jaxtyping import Array, ArrayLike, Bool, Float
from scipy.stats import poisson

from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _ppoisson(x: Union[Float, ArrayLike], rate: Union[Float, ArrayLike]) -> Array:
    k = jnp.floor(x) + 1.0
    return 1 - gammainc(k, rate)


def ppoisson(
    q: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the cumulative distribution function of the Poisson distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The cumulative distribution function of the Poisson distribution evaluated at `q`.

    Example:
        >>> ppoisson(1.0, rate=1.0)
    """
    q, dtype = _promote_dtype_to_floating(q, dtype)
    q = _check_clip_distribution_domain(q, lower=0.0)
    p = svmap_(_ppoisson, q, rate)
    p = _post_process(p, lower_tail=lower_tail, log_prob=log_prob)
    return p


@filter_jit
def _dpoisson(x, rate):
    e = jnp.exp(-rate)
    numerator = rate**x
    denominator = jnp.exp(gammaln(x + 1))
    return (numerator / denominator) * e


def dpoisson(
    x: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the probability density function of the Poisson distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the PDF.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the PDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The probability density function of the Poisson distribution evaluated at `x`.

    Examples:
        >>> dpoisson(1.0, rate=1.0)
    """
    x, _ = _promote_dtype_to_floating(x, dtype)
    x = _check_clip_distribution_domain(x, lower=0.0)
    grads = svmap_(_dpoisson, x, rate)
    grads = _post_process(grads, lower_tail=lower_tail, log_prob=log_prob)
    return grads


@filter_jit
def _rpoisson(
    key: KeyArray,
    rate: Union[Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype=jnp.int_,
):
    return jrand.poisson(key, rate, shape=sample_shape, dtype=dtype)


def rpoisson(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    rate: Union[Float, ArrayLike] = None,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.int_,
) -> ArrayLike:
    """Generates samples from the Poisson distribution.

    Args:
        key (KeyArray): Random number generator state used for random sampling.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        sample_shape (Optional[Shape], optional): Shape of the output array. Defaults to None.
        lower_tail (bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.int_.

    Returns:
        ArrayLike: Samples from the Poisson distribution.

    Examples:
        >>> key = jrand.PRNGKey(0)
        >>> rpoisson(key, rate=1.0)
    """
    rvs = _rpoisson(key, rate, sample_shape=sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail=lower_tail, log_prob=log_prob)
    return rvs


@filter_jit
def _qpoisson(q, rate, dtype):
    result_shape_type = ShapeDtypeStruct(jnp.shape(q), dtype)
    _scp_poisson_ppf = lambda q, rate: poisson(rate).ppf(q).astype(dtype)
    p = pure_callback(_scp_poisson_ppf, result_shape_type, q, rate)
    return p


def qpoisson(
    p: Union[Float, ArrayLike],
    rate: Union[Float, ArrayLike],
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.int_,
) -> ArrayLike:
    """Computes the quantile function of the Poisson distribution.

    Args:
        p (Union[Float, ArrayLike]): The probability at which to evaluate the quantile function.
        rate (Union[Float, ArrayLike]): The rate parameter of the Poisson distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the quantile function. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.int_.

    Returns:
        ArrayLike: The quantile function of the Poisson distribution evaluated at `p`.

    Example:
        >>> qpoisson(0.5, rate=1.0)
    """
    p, _ = _promote_dtype_to_floating(p, None)
    p = _check_clip_probability(p, lower_tail, log_prob)
    q = svmap_(_qpoisson, p, rate, dtype)
    return q



import jax.numpy as jnp
from jax import vmap


def cube_map(func, *inputs):

    n = len(inputs)
    args = []
    arg_sizes = []
    all_scalar = True

    for inp in inputs:
        if not jnp.isscalar(inp):
            all_scalar = False

        inp = jnp.asarray(inp)
        args.append(inp)
        arg_sizes.append(inp.size)

    if all_scalar:
        return jnp.atleast_1d(func(*args))

    else:
        for i in range(n):
            in_axes = [None] * n
            if arg_sizes[n - i - 1] > 1:
                in_axes[n - i - 1] = 0
                func = vmap(func, in_axes=in_axes)

    return func(*args)



from typing import List, NamedTuple

import jax.numpy as jnp
import pandas as pd
import patsy
from jaxtyping import Array

from ..core import make_partial_pipe


class Matrices(NamedTuple):
    y: Array
    X: Array
    y_names: List[str]
    X_names: List[str]


@make_partial_pipe
def get_design_matrices(
    data: pd.DataFrame,
    formula: str,
    eval_env=0,
    NA_action="drop",
    return_type="matrix",
    dtype: jnp.dtype = None,
) -> Matrices:
    """Get design matrices from a formula and data.

    Args:
        data: A pandas DataFrame containing the data.
        formula: A string representation of a formula.
        eval_env: An environment mapping column names to arrays which patsy can use when evaluating a formula.
        NA_action: A string specifying how to handle missing values. One of "drop", "raise", or "na_action".
        return_type: A string specifying the return type. One of "matrix", "dataframe", or "design_info".
        dtype: The data type to use for the design matrices.

    Returns:
        Matrices: A named tuple containing the design matrices for the response and predictors, as well as the names of the columns in each matrix.

    """
    y, X = patsy.dmatrices(
        formula, data, eval_env=eval_env, NA_action=NA_action, return_type=return_type
    )
    if dtype is None:
        dtype = jnp.promote_types(y.dtype, X.dtype)
    return Matrices(
        jnp.asarray(y, dtype=dtype),
        jnp.asarray(X, dtype=dtype),
        y.design_info.column_names,
        X.design_info.column_names,
    )



from typing import Optional

import jax.numpy as jnp
from equinox import filter_jit
from jax.scipy.stats import rankdata
from jaxtyping import ArrayLike


def spearmanr(x: ArrayLike, y: Optional[ArrayLike] = None, axis: int = 0) -> ArrayLike:
    """Calculates a Spearman rank-order correlation coefficient and the p-value to test for non-correlation.

    Args:
        x (ArrayLike): An array of values.
        y (Optional[ArrayLike], optional): An array of values. Defaults to None.
        axis (int, optional): The axis along which to calculate. Defaults to 0.

    Raises:
        ValueError: If the supplied axis argument is greater than 1 or if the number of dimensions of the array is greater than 2.

    Returns:
        A array-like containing the Spearman rank-order correlation coefficient and the p-value to test for non-correlation.

    Examples:
        >>> import jax.numpy as jnp
        >>> from stamox.functions import spearmanr
        >>> x = jnp.array([1, 2, 3, 4, 5])
        >>> y = jnp.array([5, 6, 7, 8, 7])
        >>> spearmanr(x, y)
        Array(0.8207823038101196, dtype=float32)
    """
    if axis is not None and axis > 1:
        raise ValueError(
            "spearmanr only handles 1-D or 2-D arrays, "
            "supplied axis argument {}, please use only "
            "values 0, 1 or None for axis".format(axis)
        )
    x = jnp.asarray(x)

    if axis is None:
        x = x.ravel()
        axis_out = 0
    else:
        axis_out = axis

    if x.ndim > 2:
        raise ValueError("spearmanr only handles 1-D or 2-D arrays")

    if y is None:
        if x.ndim < 2:
            raise ValueError("`spearmanr` needs at least 2 " "variables to compare")
    else:
        y = jnp.asarray(y, dtype=x.dtype)
        if axis is None:
            y = y.ravel()
        if axis_out == 0:
            x = jnp.column_stack((x, y))
        else:
            x = jnp.row_stack((x, y))
    return _spearmanr(x, axis_out)


@filter_jit
def _spearmanr(x, axis):
    x_ranked = jnp.apply_along_axis(rankdata, axis, x)
    coef = jnp.corrcoef(x_ranked, rowvar=axis)
    if coef.shape == (2, 2):
        coef = coef[1, 0]
    return coef



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Bool, Float

from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _ppareto(
    x: Union[Float, ArrayLike],
    scale: Union[Float, ArrayLike],
    alpha: Union[Float, ArrayLike],
):
    return 1 - jnp.power(scale / x, alpha)


def ppareto(
    q: Union[Float, ArrayLike],
    scale: Union[Float, ArrayLike],
    alpha: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the cumulative distribution function of the Pareto distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        scale (Union[Float, ArrayLike]): The scale parameter of the Pareto distribution.
        alpha (Union[Float, ArrayLike]): The shape parameter of the Pareto distribution.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The cumulative distribution function of the Pareto distribution evaluated at `q`.

    Example:
        >>> ppareto(0.2, 0.1, 2.0)
        Array(0.75, dtype=float32, weak_type=True)
    """
    q, _ = _promote_dtype_to_floating(q, dtype)
    q = svmap_(_check_clip_distribution_domain, q, scale)
    p = svmap_(_ppareto, q, scale, alpha)
    p = _post_process(p, lower_tail, log_prob)
    return p


_dpareto = filter_grad(filter_jit(_ppareto))


def dpareto(
    x: Union[Float, ArrayLike],
    scale: Union[Float, ArrayLike],
    alpha: Union[Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the density of the Pareto distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the density.
        scale (Union[Float, ArrayLike]): The scale parameter of the Pareto distribution.
        alpha (Union[Float, ArrayLike]): The shape parameter of the Pareto distribution.
        lower_tail (bool, optional): Whether to compute the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The density of the Pareto distribution evaluated at `x`.

    Example:
        >>> dpareto(0.2, 0.1, 2.0)
        Array([2.4999998], dtype=float32, weak_type=True)
    """
    x, _ = _promote_dtype_to_floating(x, dtype)
    grads = svmap_(_dpareto, x, scale, alpha)
    grads = _post_process(grads, lower_tail, log_prob)
    return grads


@filter_jit
def _qpareto(
    q: Union[Float, ArrayLike],
    scale: Union[Float, ArrayLike],
    alpha: Union[Float, ArrayLike],
):
    return scale / jnp.power(1 - q, 1 / alpha)


def qpareto(
    p: Union[Float, ArrayLike],
    scale: Union[Float, ArrayLike],
    alpha: Union[Float, ArrayLike],
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the quantile function of the Pareto distribution.

    Args:
        p (Union[Float, ArrayLike]): Quantiles to compute.
        scale (Union[Float, ArrayLike]): Scale parameter of the Pareto distribution.
        alpha (Union[Float, ArrayLike]): Shape parameter of the Pareto distribution.
        lower_tail (Bool, optional): Whether to compute the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The quantiles of the Pareto distribution.

    Example:
        >>> qpareto(0.2, 0.1, 2.0)
        Array([0.1118034], dtype=float32, weak_type=True)
    """
    p, _ = _promote_dtype_to_floating(p, dtype)
    p = _check_clip_probability(p, lower_tail, log_prob)
    return svmap_(_qpareto, p, scale, alpha)


@filter_jit
def _rpareto(
    key: KeyArray,
    scale: Union[Float, ArrayLike],
    alpha: Union[Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(scale), jnp.shape(alpha))
    scale = jnp.broadcast_to(scale, sample_shape)
    alpha = jnp.broadcast_to(alpha, sample_shape)
    return jrand.pareto(key, alpha, shape=sample_shape, dtype=dtype) * scale


def rpareto(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    scale: Union[Float, ArrayLike] = None,
    alpha: Union[Float, ArrayLike] = None,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generate random variable following a Pareto distribution.

    Args:
        key (KeyArray): A random number generator key.
        sample_shape (Optional[Shape], optional): The shape of the samples to be drawn. Defaults to None.
        scale (Union[Float, ArrayLike]): The scale parameter of the Pareto distribution.
        alpha (Union[Float, ArrayLike]): The shape parameter of the Pareto distribution.
        lower_tail (Bool, optional): Whether to calculate the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: random variable following a Pareto distribution.

    Example:
        >>> rpareto(jax.random.PRNGKey(0), sample_shape=(2, 3), scale=0.1, alpha=2.0)
        Array([[0.15330292, 0.10539087, 0.19686179],
                [0.30740616, 0.15743963, 0.13524036]], dtype=float32)
    """
    rvs = _rpareto(key, scale, alpha, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs



from functools import partial, wraps
from typing import (
    Any,
    Callable,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import equinox as eqx

from .base import Functional


T = TypeVar("T")


class Pipe(eqx.Module):
    """A class for creating a pipe of functions.

    Attributes:
        funcs (Tuple[Functional, ...]): A tuple of Functional objects.
    """

    funcs: Tuple[Functional, ...]

    def __init__(self, funcs: Sequence[Functional]) -> None:
        """Initialize the Pipe object.

        Args:
            funcs (Sequence[Functional]): A sequence of Functional objects.
        """
        self.funcs = tuple(funcs)

    def __call__(self, x: Any = None, *args, **kwargs):
        """Call the Pipe object.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Any: The output of the Pipe object.
        """
        for fn in self.funcs:
            x = fn(x, *args, **kwargs)
        return x

    def __getitem__(self, i: Union[int, slice, str]) -> Functional:
        """Get an item from the Pipe object.

        Args:
            i (Union[int, slice, str]): The index or name of the item.

        Returns:
            Functional: The item at the given index or with the given name.

        Raises:
            TypeError: If the type of the index is not supported.
            ValueError: If no function names the given string.
        """
        if isinstance(i, int):
            return self.funcs[i]
        elif isinstance(i, slice):
            return Pipe(self.funcs[i])
        elif isinstance(i, str):
            _f = []
            i = i.lower()
            for f in self.funcs:
                if f.name.lower() == i:
                    _f.append(f)

            if len(_f) == 0:
                raise ValueError(f"No Function Names {i}")
            if len(_f) > 1:
                return Pipe(_f)
            else:
                return _f[0]
        else:
            raise TypeError(f"Indexing with type {type(i)} is not supported")

    def __iter__(self) -> Functional:
        """Iterate over the Pipe object.

        Yields:
            Functional: The next item in the Pipe object.
        """
        yield from self.funcs

    def __len__(self) -> int:
        """Get the length of the Pipe object.

        Returns:
            int: The length of the Pipe object.
        """
        return len(self.funcs)

    def __rshift__(self, _next):
        """Create a new Pipe object by appending another item.

        Args:
            _next (Any): The item to append.

        Returns:
            Pipe: A new Pipe object with the given item appended.
        """
        if not isinstance(_next, Functional):
            _next = Functional(fn=_next)
        if isinstance(_next, Pipe):
            return Pipe([*self.funcs, *_next])
        return Pipe([*self.funcs, _next])


class Pipeable(Functional):
    """A class for pipeable functions.

    Attributes:
        value (Any): The value to be piped.
    """

    value: Any

    def __init__(self, *args, **kwargs):
        super().__init__(name="PipeableData", fn=None)
        """Initialize the Pipeable object.

        Args:
            value (Any): The value to be piped.
        """
        # make weakref for large scale data
        self.value = tuple(args) if len(args) > 1 else args[0]

    def __call__(self, *args, **kwargs):
        """Pipe the value through the function.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Any: The piped value.
        """
        return self.value


def make_pipe(
    func: Optional[Callable[..., T]] = None, name: str = None
) -> Callable[..., T]:
    """Makes a Function Pipeable.

    Args:
        func (Callable): Function or Callable Class.
        name (str, optional): Name of the Function. Defaults to "PipeableFunc".
        kwargs (optional): Additional keyword arguments.

    Returns:
        Callable: The wrapped function.

    Examples:
        >>> from stamox import make_pipe
        >>> @make_pipe
        ... def add(x):
        ...     return x + 1
        >>> h = add >> add >> add
        >>> h(1)
        4
    """

    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable[..., T]:
        if isinstance(func, Functional):
            if func.func is not None:
                func = func.func

        functional = Functional(name=name, fn=func)
        return functional

    return wrap if func is None else wrap(func)


def make_partial_pipe(
    func: Optional[Callable[..., T]] = None, name: str = None
) -> Callable[..., T]:
    """Makes a Partial Function Pipe.

    Args:
        func (Callable): Function or Callable Class.
        name (str, optional): Name of the Function. Defaults to "PipeableFunc".
        kwargs (dict): Keyword arguments for the function.

    Returns:
        Callable: A partial function pipe.

    Examples:
        >>> from stamox import make_partial_pipe
        >>> @make_partial_pipe
        ... def add(x, y):
        ...     return x + y
        >>> h = add(y=1) >> add(y=2) >> add(y=3)
        >>> h(1)
        7
    """
    if name is None and func is not None:
        if hasattr(func, "name"):
            name = func.name
        elif hasattr(func, "__name__"):
            name = func.__name__
        else:
            name = "none"

    @wraps(func)
    def wrap(func: Callable[..., T]) -> Callable:
        if isinstance(func, Functional):
            if func.func is not None:
                if func.pipe_type == "pmap" or func.pipe_type == "vmap":
                    raise ValueError(
                        "You can not use pipe_pmap or piep_vmap with make_partial_pipe, use make_pipe or pipe_* instead."
                    )
                func = func.func

        @wraps(func)
        def partial_fn(*args, **kwargs):
            if len(args) != 0:
                return func(*args, **kwargs)
            fn = partial(func, **kwargs)
            return Functional(name=name, fn=fn)

        return partial_fn

    return wrap if func is None else wrap(func)



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit
from jax import lax
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Bool, Float

from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _pcauchy(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
) -> ArrayLike:
    dtype = lax.dtype(x)
    loc = jnp.asarray(loc, dtype=dtype)
    scale = jnp.asarray(scale, dtype=dtype)
    scaled = (x - loc) / scale
    return jnp.arctan(scaled) / jnp.pi + 0.5


def pcauchy(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the cumulative denisty probability c function of the Cauchy distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        loc (Union[Float, ArrayLike], optional): The location parameter of the Cauchy distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Cauchy distribution. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The cumulative density function of the Cauchy distribution.

    Example:
        >>> pcauchy(1.0, loc=0.0, scale=1.0, lower_tail=True, log_prob=False)
        Array(0.75, dtype=float32, weak_type=True)
    """
    q, dtype = _promote_dtype_to_floating(q, dtype)
    q = _check_clip_distribution_domain(q)
    p = svmap_(_pcauchy, q, loc, scale)
    p = _post_process(p, lower_tail, log_prob)
    return p


_dcauchy = filter_grad(filter_jit(_pcauchy))


def dcauchy(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the pdf of the Cauchy distribution.

    Args:
        x (Union[Float, ArrayLike]): The input values.
        loc (Union[Float, ArrayLike], optional): The location parameter. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The pdf of the Cauchy distribution.

    Example:
        >>> dcauchy(1.0, loc=0.0, scale=1.0)
        Array([0.15915494], dtype=float32, weak_type=True)
    """
    x, dtype = _promote_dtype_to_floating(x, dtype)
    x = _check_clip_distribution_domain(x)
    grads = svmap_(_dcauchy, x, loc, scale)
    grads = _post_process(grads, lower_tail, log_prob)
    return grads


@filter_jit
def _qcauchy(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    dtype = lax.dtype(q)
    loc = jnp.asarray(loc, dtype=dtype)
    scale = jnp.asarray(scale, dtype=dtype)
    return lax.add(loc, lax.mul(scale, lax.tan(lax.mul(jnp.pi, lax.sub(q, 0.5)))))


def qcauchy(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the quantile of the Cauchy distribution.

    Args:
        q (Union[float, array-like]): Quantiles to compute.
        loc (Union[float, array-like], optional): Location parameter. Defaults to 0.0.
        scale (Union[float, array-like], optional): Scale parameter. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The quantiles of the Cauchy distribution.

    Example:
        >>> qcauchy(0.5, loc=1.0, scale=2.0)
        Array([1.], dtype=float32, weak_type=True)
    """
    q, dtype = _promote_dtype_to_floating(q, dtype)
    q = _check_clip_probability(q, lower_tail, log_prob)
    return svmap_(_qcauchy, q, loc, scale)


@filter_jit
def _rcauchy(
    key: KeyArray,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    loc = jnp.broadcast_to(loc, sample_shape)
    scale = jnp.broadcast_to(scale, sample_shape)
    loc = jnp.asarray(loc, dtype=dtype)
    scale = jnp.asarray(scale, dtype=dtype)
    return jrand.cauchy(key, sample_shape, dtype=dtype) * scale + loc


def rcauchy(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generates random samples from the Cauchy distribution.

    Args:
        key: A PRNGKey to use for generating the samples.
        sample_shape: The shape of the output array.
        loc: The location parameter of the Cauchy distribution.
        scale: The scale parameter of the Cauchy distribution.
        lower_tail: Whether to return the lower tail probability.
        log_prob: Whether to return the log probability.
        dtype: The dtype of the output.

    Returns:
        ArrayLike: An array of samples from the Cauchy distribution.

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> rcauchy(key, sample_shape=(2, 3), loc=0.0, scale=1.0)
        Array([[ 0.23841971, -3.0880406 ,  0.9507532 ],
                [ 2.8963416 ,  0.31303588, -0.14792857]], dtype=float32)
    """
    rvs = _rcauchy(key, loc, scale, sample_shape, dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs



from jax import lax
from jax.scipy.special import betainc
from tensorflow_probability.substrates.jax.math import special as tfp_special


def fdtrc(a, b, x):
    dtype = lax.dtype(x)
    a = lax.convert_element_type(a, dtype)
    b = lax.convert_element_type(b, dtype)
    w = lax.div(b, lax.add(b, lax.mul(a, x)))
    half = lax.convert_element_type(0.5, w.dtype)
    a = lax.mul(half, a)
    b = lax.mul(half, b)
    return betainc(a, b, w)


def fdtr(a, b, x):
    dtype = lax.dtype(x)
    a = lax.convert_element_type(a, dtype)
    b = lax.convert_element_type(b, dtype)
    w = lax.mul(a, x)
    w = lax.div(w, lax.add(b, w))
    half = lax.convert_element_type(0.5, dtype)
    a = lax.mul(half, a)
    b = lax.mul(half, b)
    return betainc(a, b, w)


def fdtri(a, b, y):
    dtype = lax.dtype(y)
    one = lax.convert_element_type(1.0, dtype)
    half = lax.convert_element_type(0.5, dtype)
    eps = lax.convert_element_type(1e-3, dtype)
    a = lax.convert_element_type(a, dtype)
    b = lax.convert_element_type(b, dtype)
    y = lax.sub(one, y)
    a = lax.mul(half, a)
    b = lax.mul(half, b)
    w = betainc(a, b, half)
    cond0 = (w > y) | (y < eps)
    w = lax.select(
        cond0,
        tfp_special.betaincinv(b, a, y),
        tfp_special.betaincinv(a, b, lax.sub(one, y)),
    )
    left_out = lax.div(lax.sub(b, lax.mul(b, w)), lax.mul(a, w))
    right_out = lax.div(lax.mul(b, w), lax.mul(a, lax.sub(one, w)))
    x = lax.select(cond0, left_out, right_out)
    return x



import jax.numpy as jnp
from jax import lax
from jaxtyping import ArrayLike


def boxcox(x: ArrayLike, lmbda: ArrayLike, dtype=jnp.float_) -> ArrayLike:
    """Computes the Box-Cox transformation of a given array.

    Args:
        x: An array-like object to be transformed.
        lmbda: An array-like object containing the lambda values for the transformation.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        The boxcox transformed array.

    Example:
        >>> from stamox.transformation import boxcox
        >>> import jax.numpy as jnp
        >>> x = jnp.array([1, 2, 3, 4, 5], dtype=jnp.float32)
        >>> lmbda = jnp.array([0, 0, 0, 0, 0], dtype=jnp.float32)
        >>> boxcox(x, lmbda)
        Array([0.        , 0.6931472 , 1.0986123 , 1.3862944 , 1.6094378 ], dtype=float32)
    """
    if dtype is None:
        dtype = lax.dtype(x)
    x = jnp.asarray(x, dtype)
    lmbda = jnp.asarray(lmbda, dtype=dtype)
    return lax.select(lmbda == 0, jnp.log(x), (jnp.power(x, lmbda) - 1) / lmbda)


def z_fisher(rho: ArrayLike, dtype=jnp.float_) -> ArrayLike:
    """Computes the Fisher z-transform of a given array.

    Args:
        rho: An array-like object to be transformed.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        The Fisher z-transformed array.

    Example:
        >>> from stamox.functions import z_fisher
        >>> import jax.numpy as jnp
        >>> rho = jnp.array([0.1, 0.2, 0.3, 0.4, 0.5], dtype=jnp.float32)
        >>> z_fisher(rho)
        Array([0.10033537, 0.2013589 , 0.30469212, 0.41073018, 0.51991177], dtype=float32)
    """
    if dtype is None:
        dtype = lax.dtype(rho)
    rho = jnp.asarray(rho, dtype=dtype)
    rho = jnp.clip(rho, -1.0, 1.0)
    return jnp.arctanh(rho)



import jax.numpy as jnp

from .anova import one_way
from .basic import scale
from .cluster import kmeans
from .core import (
    make_partial_pipe,
    make_pipe,
    predict,
    summary,
)
from .correlation import cor, pearsonr, spearmanr
from .decomposition import princomp
from .distribution import (
    dbeta,
    dbinom,
    dcauchy,
    dchisq,
    dexp,
    dF,
    dgamma,
    dlaplace,
    dnorm,
    dpareto,
    dpoisson,
    dt,
    dunif,
    dweibull,
    ecdf,
    pbeta,
    pbinom,
    pcauchy,
    pchisq,
    pexp,
    pF,
    pgamma,
    plaplace,
    pnorm,
    ppareto,
    ppoisson,
    pt,
    punif,
    pweibull,
    qbeta,
    qbinom,
    qcauchy,
    qchisq,
    qexp,
    qF,
    qgamma,
    qlaplace,
    qnorm,
    qpareto,
    qpoisson,
    qt,
    qunif,
    qweibull,
    rbeta,
    rbinom,
    rcauchy,
    rchisq,
    rexp,
    rF,
    rgamma,
    rlaplace,
    rnorm,
    rpareto,
    rpoisson,
    rt,
    runif,
    rweibull,
)
from .hypothesis import (
    bartlett_test,
    durbin_watson_test,
    friedman_test,
)
from .regression import lm
from .sample import bootstrap, bootstrap_sample, jackknife, jackknife_sample
from .transformation import boxcox, z_fisher


one_way = make_partial_pipe(one_way, name="one_way")
kmeans = make_partial_pipe(kmeans, name="kmeans")
spearmanr = make_partial_pipe(spearmanr, name="spearmanr")
pearsonr = make_partial_pipe(pearsonr, name="pearsonr")
cor = make_partial_pipe(cor, name="cor")
bartlett_test = make_partial_pipe(bartlett_test, name="bartlett_test")
durbin_watson_test = make_partial_pipe(durbin_watson_test, name="durbin_watson_test")
friedman_test = make_partial_pipe(friedman_test, name="friedman_test")
bootstrap = make_partial_pipe(bootstrap, name="bootstrap")
bootstrap_sample = make_partial_pipe(bootstrap_sample, name="bootstrap_sample")
jackknife = make_partial_pipe(jackknife, name="jackknife")
jackknife_sample = make_pipe(jackknife_sample, name="jackknife_sample")
boxcox = make_partial_pipe(boxcox, name="boxcox")
z_fisher = make_partial_pipe(z_fisher, name="z_fisher")
lm = make_partial_pipe(lm, name="lm")
dbeta = make_partial_pipe(dbeta, name="dbeta")
dbinom = make_partial_pipe(dbinom, name="dbinom")
dcauchy = make_partial_pipe(dcauchy, name="dcauchy")
dchisq = make_partial_pipe(dchisq, name="dchisq")
dexp = make_partial_pipe(dexp, name="dexp")
dF = make_partial_pipe(dF, name="dF")
dgamma = make_partial_pipe(dgamma, name="dgamma")
dlaplace = make_partial_pipe(dlaplace, name="dlaplace")
dnorm = make_partial_pipe(dnorm, name="dnorm")
dpareto = make_partial_pipe(dpareto, name="dpareto")
dpoisson = make_partial_pipe(dpoisson, name="dpoisson")
dt = make_partial_pipe(dt, name="dt")
dunif = make_partial_pipe(dunif, name="dunif")
dweibull = make_partial_pipe(dweibull, name="dweibull")
ecdf = make_partial_pipe(ecdf, name="ecdf")
pbeta = make_partial_pipe(pbeta, name="pbeta")
pbinom = make_partial_pipe(pbinom, name="pbinom")
pcauchy = make_partial_pipe(pcauchy, name="pcauchy")
pchisq = make_partial_pipe(pchisq, name="pchisq")
pexp = make_partial_pipe(pexp, name="pexp")
pF = make_partial_pipe(pF, name="pF")
pgamma = make_partial_pipe(pgamma, name="pgamma")
plaplace = make_partial_pipe(plaplace, name="plaplace")
pnorm = make_partial_pipe(pnorm, name="pnorm")
ppareto = make_partial_pipe(ppareto, name="ppareto")
ppoisson = make_partial_pipe(ppoisson, name="ppoisson")
pt = make_partial_pipe(pt, name="pt")
punif = make_partial_pipe(punif, name="punif")
pweibull = make_partial_pipe(pweibull, name="pweibull")
qbeta = make_partial_pipe(qbeta, name="qbeta")
qbinom = make_partial_pipe(qbinom, name="qbinom")
qcauchy = make_partial_pipe(qcauchy, name="qcauchy")
qchisq = make_partial_pipe(qchisq, name="qchisq")
qexp = make_partial_pipe(qexp, name="qexp")
qF = make_partial_pipe(qF, name="qF")
qgamma = make_partial_pipe(qgamma, name="qgamma")
qlaplace = make_partial_pipe(qlaplace, name="qlaplace")
qnorm = make_partial_pipe(qnorm, name="qnorm")
qpareto = make_partial_pipe(qpareto, name="qpareto")
qpoisson = make_partial_pipe(qpoisson, name="qpoisson")
qt = make_partial_pipe(qt, name="qt")
qunif = make_partial_pipe(qunif, name="qunif")
qweibull = make_partial_pipe(qweibull, name="qweibull")
rbeta = make_partial_pipe(rbeta, name="rbeta")
rbinom = make_partial_pipe(rbinom, name="rbinom")
rcauchy = make_partial_pipe(rcauchy, name="rcauchy")
rchisq = make_partial_pipe(rchisq, name="rchisq")
rexp = make_partial_pipe(rexp, name="rexp")
rF = make_partial_pipe(rF, name="rF")
rgamma = make_partial_pipe(rgamma, name="rgamma")
rlaplace = make_partial_pipe(rlaplace, name="rlaplace")
rnorm = make_partial_pipe(rnorm, name="rnorm")
rpareto = make_partial_pipe(rpareto, name="rpareto")
rpoisson = make_partial_pipe(rpoisson, name="rpoisson")
rt = make_partial_pipe(rt, name="rt")
runif = make_partial_pipe(runif, name="runif")
rweibull = make_partial_pipe(rweibull, name="rweibull")
predict = make_partial_pipe(predict, name="predict")
summary = make_partial_pipe(summary, name="summary")
mean = make_partial_pipe(jnp.mean, "mean")
sd = make_partial_pipe(jnp.std, "sd")
var = make_partial_pipe(jnp.var, "var")
median = make_partial_pipe(jnp.median, "median")
quantile = make_partial_pipe(jnp.quantile, "quantile")
min = make_partial_pipe(jnp.min, "min")
max = make_partial_pipe(jnp.max, "max")
sum = make_partial_pipe(jnp.sum, "sum")
prod = make_partial_pipe(jnp.prod, "prod")
cumsum = make_partial_pipe(jnp.cumsum, "cumsum")
cumprod = make_partial_pipe(jnp.cumprod, "cumprod")
diff = make_partial_pipe(jnp.diff, "diff")
scale = make_partial_pipe(scale, "scale")
princomp = make_partial_pipe(princomp, "princomp")

__all__ = [
    "one_way",
    "kmeans",
    "spearmanr",
    "pearsonr",
    "cor",
    "bartlett_test",
    "durbin_watson_test",
    "friedman_test",
    "lm",
    "bootstrap",
    "bootstrap_sample",
    "jackknife",
    "jackknife_sample",
    "boxcox",
    "z_fisher",
    "dbeta",
    "dbinom",
    "dcauchy",
    "dchisq",
    "dexp",
    "dF",
    "dgamma",
    "dlaplace",
    "dnorm",
    "dpareto",
    "dpoisson",
    "dt",
    "dunif",
    "dweibull",
    "ecdf",
    "pbeta",
    "pbinom",
    "pcauchy",
    "pchisq",
    "pexp",
    "pF",
    "pgamma",
    "plaplace",
    "pnorm",
    "ppareto",
    "ppoisson",
    "pt",
    "punif",
    "pweibull",
    "qbeta",
    "qbinom",
    "qcauchy",
    "qchisq",
    "qexp",
    "qF",
    "qgamma",
    "qlaplace",
    "qnorm",
    "qpareto",
    "qpoisson",
    "qt",
    "qunif",
    "qweibull",
    "rbeta",
    "rbinom",
    "rcauchy",
    "rchisq",
    "rexp",
    "rF",
    "rgamma",
    "rlaplace",
    "rnorm",
    "rpareto",
    "rpoisson",
    "rt",
    "runif",
    "rweibull",
    "predict",
    "summary",
    "mean",
    "sd",
    "var",
    "median",
    "quantile",
    "min",
    "max",
    "sum",
    "prod",
    "cumsum",
    "cumprod",
    "diff",
    "scale",
    "princomp",
]



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit, filter_vmap
from jax import lax
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Float, Int
from tensorflow_probability.substrates.jax.math import special as tfp_special

from ._utils import (
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _pt(
    x: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    df = lax.convert_element_type(df, x.dtype)
    loc = lax.convert_element_type(loc, x.dtype)
    scale = lax.convert_element_type(scale, x.dtype)
    scaled = lax.div(lax.sub(x, loc), scale)
    scaled_squared = lax.integer_pow(scaled, 2)
    beta_value = lax.div(df, lax.add(df, scaled_squared))
    return 0.5 * (
        1.0
        + jnp.sign(scaled)
        - jnp.sign(scaled) * tfp_special.betainc(0.5 * df, 0.5, beta_value)
    )


def pt(
    q: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the probability of a given value for Student T distribution.

    Args:
        q: The value to calculate the probability of.
        df: The degrees of freedom of the distribution.
        loc: The location parameter of the distribution.
        scale: The scale parameter of the distribution.
        lower_tail: Whether to calculate the lower tail probability or not.
        log_prob: Whether to return the log probability or not.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The cdf value of the given value for Student T distribution.

    Example:
        >>> pt(1.0, 1.0)
        Array(0.74999994, dtype=float32, weak_type=True)
    """
    q, _ = _promote_dtype_to_floating(q, dtype)
    p = svmap_(_pt, q, df, loc, scale)
    p = _post_process(p, lower_tail, log_prob)
    return p


_dt = filter_jit(filter_grad(_pt))


def dt(
    x: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the probability density function of a Student's t-distribution.

    Args:
        x: A float or array-like object representing the values at which to evaluate
        the probability density function.
        df: Degrees of freedom for the Student's t-distribution.
        loc: Location parameter for the Student's t-distribution. Defaults to 0.0.
        scale: Scale parameter for the Student's t-distribution. Defaults to 1.0.
        lower_tail: Whether to return the lower tail probability. Defaults to True.
        log_prob: Whether to return the log probability. Defaults to False.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The probability density function evaluated at `x`.

    Example:
        >>> dt(1.0, 1.0)
        Array(0.1591549, dtype=float32, weak_type=True)
    """
    x, _ = _promote_dtype_to_floating(x, dtype)
    grads = svmap_(_dt, x, df, loc, scale)
    grads = _post_process(grads, lower_tail, log_prob)
    return grads


@filter_jit
def _qt(
    q: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    df = lax.convert_element_type(df, q.dtype)
    loc = lax.convert_element_type(loc, q.dtype)
    scale = lax.convert_element_type(scale, q.dtype)
    one = lax.convert_element_type(1, q.dtype)
    beta_value = tfp_special.betaincinv(0.5 * df, 0.5, 1 - jnp.abs(1 - 2 * q))
    scaled_squared = df * (one / beta_value - one)
    scaled = jnp.sign(q - 0.5) * jnp.sqrt(scaled_squared)
    return scaled * scale + loc


def qt(
    p: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the quantile of Student T distribution.

    Args:
        p: A float or array-like object representing the quantile to be calculated.
        df: An int, float, or array-like object representing the degrees of freedom.
        loc: An optional float or array-like object representing the location parameter. Defaults to 0.0.
        scale: An optional float or array-like object representing the scale parameter. Defaults to 1.0.
        lower_tail: A boolean indicating whether the lower tail should be used. Defaults to True.
        log_prob: A boolean indicating whether the probability should be logged. Defaults to False.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The quantile of the Student T distribution.

    Example:
        >>> qt(0.5, 1.0)
        Array(0., dtype=float32, weak_type=True)
    """
    p, _ = _promote_dtype_to_floating(p, dtype)
    p = _check_clip_probability(p, lower_tail=lower_tail, log_prob=log_prob)
    q = svmap_(_qt, p, df, loc, scale)
    return q


@filter_jit
def _rt(
    key: KeyArray,
    df: Union[Int, Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(
            jnp.shape(df), jnp.shape(loc), jnp.shape(scale)
        )
    scale = jnp.broadcast_to(scale, sample_shape)
    loc = jnp.broadcast_to(loc, sample_shape)
    return jrand.t(key, df, sample_shape, dtype=dtype) * scale + loc


def rt(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    df: Union[Int, Float, ArrayLike] = None,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generates random numbers from a t-distribution.

    Args:
        key: Type of the random number generator.
        sample_shape: Shape of the output array.
        df: Degrees of freedom.
        loc: Location parameter.
        scale: Scale parameter.
        lower_tail: Whether to return the lower tail probability. Defaults to True.
        log_prob: Whether to return the log probability. Defaults to False.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: Random numbers from a t-distribution.

    Example:
        >>> rt(key, (2, 3), 1.0)
        Array([[1.9982358e+02, 2.3699088e-01, 6.6509140e-01],
                [5.3681795e-02, 3.3967651e+01, 6.8611817e+00]], dtype=float32)
    """
    rvs = _rt(key, df, loc, scale, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit
from jax import lax
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Bool, Float

from ._utils import (
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _punif(
    x: Union[Float, ArrayLike],
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
):
    mini = lax.convert_element_type(mini, x.dtype)
    maxi = lax.convert_element_type(maxi, x.dtype)
    p = lax.div(lax.sub(x, mini), lax.sub(maxi, mini))
    p = jnp.clip(p, a_min=mini, a_max=maxi)
    return p


def punif(
    q: Union[Float, ArrayLike],
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the cumulative distribution function of the uniform distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        mini (Union[Float, ArrayLike], optional): The minimum value of the uniform distribution. Defaults to 0.0.
        maxi (Union[Float, ArrayLike], optional): The maximum value of the uniform distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The cumulative distribution function of the uniform distribution evaluated at `q`.

    Example:
        >>> punif(0.5)
        Array(0.5, dtype=float32, weak_type=True)
    """
    q, _ = _promote_dtype_to_floating(q, dtype)
    p = svmap_(_punif, q, mini, maxi)
    p = _post_process(p, lower_tail, log_prob)
    return p


_dunif = filter_grad(filter_jit(_punif))


def dunif(
    x: Union[Float, ArrayLike],
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the probability density function of a uniform distribution.

    Args:
        x (Union[Float, ArrayLike]): The value or array of values for which to calculate the probability density.
        mini (Union[Float, Array], optional): The lower bound of the uniform distribution. Defaults to 0.0.
        maxi (Union[Float, Array], optional): The upper bound of the uniform distribution. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to calculate the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The probability density of the given value(s).

    Example:
        >>> dunif(0.5)
        Array(1., dtype=float32, weak_type=True)
    """
    x, _ = _promote_dtype_to_floating(x, dtype)
    p = svmap_(_dunif, x, mini, maxi)
    p = _post_process(p, lower_tail, log_prob)
    return p


@filter_jit
def _qunif(
    q: Union[Float, ArrayLike],
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
):
    x = q * (maxi - mini) + mini
    x = jnp.clip(x, a_min=mini, a_max=maxi)
    return x


def qunif(
    p: Union[Float, ArrayLike],
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """
    Computes the quantile function of a uniform distribution.

    Args:
        p (Union[Float, ArrayLike]): Quantiles to compute.
        mini (Union[Float, Array], optional): Lower bound of the uniform distribution. Defaults to 0.0.
        maxi (Union[Float, Array], optional): Upper bound of the uniform distribution. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to compute the lower tail or not. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability or not. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
       ArrayLike: The quantiles of the uniform distribution.

    Example:
        >>> qunif(0.5)
        Array(0.5, dtype=float32, weak_type=True)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = _check_clip_probability(p, lower_tail, log_prob)
    q = svmap_(_qunif, p, mini, maxi)
    return q


@filter_jit
def _runif(
    key: KeyArray,
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    return jrand.uniform(key, sample_shape, minval=mini, maxval=maxi, dtype=dtype)


def runif(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    mini: Union[Float, ArrayLike] = 0.0,
    maxi: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generates random numbers from a uniform distribution.

    Args:
        key: A PRNGKey to use for generating the random numbers.
        sample_shape: The shape of the output array.
        mini: The minimum value of the uniform distribution.
        maxi: The maximum value of the uniform distribution.
        lower_tail: Whether to generate values from the lower tail of the distribution.
        log_prob: Whether to return the log probability of the generated values.
        dtype: The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: An array of random numbers from a uniform distribution.

    Example:
        >>> runif(key, sample_shape=(2, 3))
        Array([[0.57450044, 0.09968603, 0.7419659 ],
                [0.8941783 , 0.59656656, 0.45325184]], dtype=float32)

    """
    rvs = _runif(key, mini, maxi, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs



from typing import Optional

from jaxtyping import ArrayLike

from ._pearson import pearsonr
from ._spearman import spearmanr


def cor(
    x: ArrayLike,
    y: Optional[ArrayLike] = None,
    axis: int = 0,
    method: str = "pearson",
) -> ArrayLike:
    """Calculates correlation between two arrays.

    Args:
        x (ArrayLike): The first array.
        y (Optional[ArrayLike], optional): The second array. Defaults to None.
        axis (int, optional): Axis along which the correlation is calculated. Defaults to 0.
        method (str, optional): Method used for calculating correlation. Defaults to "pearson".

    Returns:
        ArrayLike: Correlation between two arrays.

    Raises:
        NotImplementedError: If the specified method is not supported.

    Examples:
        >>> import jax.numpy as jnp
        >>> from stamox.functions import cor
        >>> x = jnp.array([1, 2, 3, 4, 5])
        >>> y = jnp.array([5, 6, 7, 8, 7])
        >>> cor(x, y)
        Array(0.8320503, dtype=float32)
    """
    if method == "pearson":
        return pearsonr(x, y, axis)
    elif method == "spearman":
        return spearmanr(x, y, axis)
    else:
        raise NotImplementedError(f"method {method} not supported")



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit
from jax import lax
from jax._src.random import Shape
from jax.random import KeyArray
from jaxtyping import ArrayLike, Bool, Float

from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _plaplace(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    dtype = lax.dtype(x)
    loc = jnp.asarray(loc, dtype=dtype)
    scale = jnp.asarray(scale, dtype=dtype)
    half = jnp.asarray(0.5, dtype=dtype)
    scaled = lax.div(lax.sub(x, loc), scale)
    subtrahend = half * lax.mul(jnp.sign(scaled), lax.expm1(-lax.abs(scaled)))
    return lax.sub(half, subtrahend)


def plaplace(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the Laplace cumulative density function.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the Plaplace PDF.
        loc (Union[Float, ArrayLike], optional): The location parameter of the Plaplace PDF. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Plaplace PDF. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail of the Plaplace PDF. Defaults to True.
        log_prob (Bool, optional): Whether to return the logarithm of the Plaplace PDF. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The Laplace CDF evaluated at `q`.

    Example:
        >>> plaplace(1.0, 1.0, 1.0)
        Array(0.5, dtype=float32, weak_type=True)
    """
    q, _ = _promote_dtype_to_floating(q, dtype)
    q = _check_clip_distribution_domain(q)
    p = svmap_(_plaplace, q, loc, scale)
    p = _post_process(p, lower_tail, log_prob)
    return p


_dlaplace = filter_grad(filter_jit(_plaplace))


def dlaplace(
    x: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the Laplace probability density function for a given x, location and scale.

    Args:
        x (Union[Float, ArrayLike]): The value at which to calculate the probability density.
        loc (Union[Float, ArrayLike], optional): The location parameter. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail of the distribution. Defaults to True.
        log_prob (Bool, optional): Whether to return the logarithm of the probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The probability density at the given x.

    Example:
        >>> dlaplace(1.0, 1.0, 1.0)
        Array(0., dtype=float32, weak_type=True)
    """
    x, _ = _promote_dtype_to_floating(x, dtype)
    x = _check_clip_distribution_domain(x)
    grads = svmap_(_dlaplace, x, loc, scale)
    grads = _post_process(grads, lower_tail, log_prob)
    return grads


@filter_jit
def _qlaplace(
    q: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    dtype = lax.dtype(q)
    loc = jnp.asarray(loc, dtype=dtype)
    scale = jnp.asarray(scale, dtype=dtype)
    half = jnp.asarray(0.5, dtype=dtype)
    a = lax.sub(half, q)
    return scale * jnp.sign(a) * jnp.log1p(-2 * jnp.abs(a)) - loc


def qlaplace(
    p: Union[Float, ArrayLike],
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the quantile of the Laplace distribution.

    Args:
        p (Union[Float, ArrayLike]): Quantiles to compute.
        loc (Union[Float, ArrayLike], optional): Location parameter. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): Scale parameter. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to compute the lower tail. Defaults to True.
        log_prob (Bool, optional): Whether to compute the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: The quantiles of the Laplace distribution.

    Example:
        >>> qlaplace(0.5, 1.0, 1.0)
        Array(1., dtype=float32, weak_type=True)
    """
    p, _ = _promote_dtype_to_floating(p, dtype)
    p = _check_clip_probability(p, lower_tail, log_prob)
    return svmap_(_qlaplace, p, loc, scale)


@filter_jit
def _rlaplace(
    key: KeyArray,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    loc = jnp.broadcast_to(loc, sample_shape)
    scale = jnp.broadcast_to(scale, sample_shape)
    return jrand.laplace(key, sample_shape, dtype=dtype) * scale + loc


def rlaplace(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    loc: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: Bool = True,
    log_prob: Bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generates random Laplace samples from a given key.

    Args:
        key (KeyArray): The PRNG key to use for generating the samples.
        sample_shape (Optional[Shape], optional): The shape of the output array. Defaults to None.
        loc (Union[Float, ArrayLike], optional): The location parameter of the Laplace distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Laplace distribution. Defaults to 1.0.
        lower_tail (Bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (Bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        ArrayLike: An array containing the random Laplace samples.

    Example:
        >>> rlaplace(key, (2, 3))
        Array([[-0.16134426,  1.6125823 , -0.6615164 ],
                [-1.5528525 , -0.21459664,  0.09816013]], dtype=float32)

    """
    rvs = _rlaplace(key, loc, scale, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs



"""Test the transformation module."""
import numpy as np
from scipy.stats import boxcox as scp_boxcox

import stamox.pipe_functions as PF
from stamox import Pipeable
from stamox.transformation import boxcox, z_fisher


np.random.seed(0)


def test_boxcox():
    x = np.random.gamma(2, 2, size=(10000, 3))
    lmbda = 2.0
    np.testing.assert_allclose(boxcox(x, lmbda), scp_boxcox(x, lmbda))


def test_partial_pipe_boxcox():
    x = np.random.gamma(2, 2, size=(10000, 3))
    lmbda = 2.0
    p = Pipeable(x) >> PF.boxcox(lmbda=2.0)
    np.testing.assert_allclose(p(x), scp_boxcox(x, lmbda))


def test_z_fisher():
    x = np.random.uniform(size=1000)
    np.testing.assert_allclose(z_fisher(x), np.arctanh(x))


def test_partial_pipe_z_fisher():
    x = np.random.uniform(size=1000)
    p = Pipeable(x) >> PF.z_fisher
    np.testing.assert_allclose(p(x), np.arctanh(x))



import functools

import jax.numpy as jnp
from equinox import filter_vmap
from jax import jit, lax


inf = jnp.inf
nan = jnp.nan


@functools.partial(jit, static_argnums=(1, 2))
def _check_clip_probability(p, lower_tail=True, log_prob=False):
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.exp(p)
    return jnp.clip(p, 0.0, 1.0)


@functools.partial(jit, static_argnums=(1, 2))
def _check_clip_distribution_domain(x, lower=-inf, upper=inf):
    return jnp.clip(x, lower, upper)


def _promote_dtype_to_floating(q, dtype):
    if dtype is None:
        if jnp.issubdtype(lax.dtype(q), jnp.integer):
            return jnp.asarray(q, dtype=jnp.float_), jnp.float_
        else:
            return jnp.asarray(q), lax.dtype(q)
    else:
        return jnp.asarray(q, dtype=dtype), dtype


def _promote_dtype_to_integer(x, dtype):
    if dtype is None:
        if jnp.issubdtype(lax.dtype(x), jnp.floating):
            return jnp.asarray(x, dtype=jnp.int_), jnp.int_
        else:
            return jnp.asarray(x), lax.dtype(x)
    else:
        return jnp.asarray(x, dtype=dtype), dtype


@functools.partial(jit, static_argnums=(1, 2))
def _post_process(p, lower_tail=True, log_prob=False):
    if not lower_tail:
        p = 1 - p
    if log_prob:
        p = jnp.log(p)
    return p


def _check_all_scalar(*args):
    for arg in args:
        if not jnp.shape(arg) == ():
            return False
    return True


def svmap_(f, *args):
    if _check_all_scalar(*args):
        return f(*args)
    else:
        return filter_vmap(f)(*args)


__all__ = [  # noqa: F405
    "inf",
    "nan",
    "_check_clip_probability",
    "_check_clip_distribution_domain",
    "_promote_dtype_to_floating",
    "_promote_dtype_to_integer",
    "_post_process",
    "_check_all_scalar",
    "svmap_",
]



import jax.numpy as jnp
import jax.random as jrandom
from equinox import filter_jit
from jax import lax, vmap
from jax._src.random import KeyArray
from jax.scipy.cluster.vq import vq
from jaxtyping import ArrayLike

from ..core import StateFunc


class KMeansState(StateFunc):
    """KMeansState class for K-means clustering.

    Attributes:
        n_clusters (int): Number of clusters.
        centers (ArrayLike): Centers of the clusters.
        cluster (ArrayLike): Cluster labels for each point.
        iters (int): Number of iterations.
        totss (float): Total sum of squares.
        betwss (float): Between sum of squares.
        withinss (float): Within sum of squares.
        tot_withinss (float): Total within sum of squares.
    """

    n_clusters: int
    centers: ArrayLike
    cluster: ArrayLike
    iters: int
    totss: float
    betwss: float
    withinss: float
    tot_withinss: float

    def __init__(
        self, n_clusters, centers, cluster, iters, totss, betwss, withinss, tot_withinss
    ):
        super().__init__(name="KMeans Cluster State", fn=None)
        self.n_clusters = n_clusters
        self.centers = centers
        self.cluster = cluster
        self.iters = iters
        self.totss = totss
        self.betwss = betwss
        self.withinss = withinss
        self.tot_withinss = tot_withinss

    def _predict(self, x: ArrayLike):
        # predict the cluster for new data
        return vq(x, self.centers)[0]


def kmeans(
    x: ArrayLike,
    n_cluster: int,
    restarts: int = 10,
    max_iters: int = 100,
    dtype: jnp.dtype = jnp.float32,
    *,
    key: KeyArray = None
):
    """Runs the K-means clustering algorithm on a given dataset.

    Args:
        x (ArrayLike): The dataset to be clustered.
        n_cluster (int): The number of clusters to generate.
        restarts (int, optional): The number of restarts for the algorithm. Defaults to 10.
        max_iters (int, optional): The maximum number of iterations for the algorithm. Defaults to 100.
        dtype (jnp.dtype, optional): The data type of the output. Defaults to jnp.float32.
        key (KeyArray, optional): A key array used for encryption. Defaults to None.

    Returns:
        KMeansState: An object containing the results of the clustering algorithm.

    Example:
        >>> from jax import random
        >>> from stamox.functions import kmeans
        >>> key = random.PRNGKey(0)
        >>> x = random.normal(key, shape=(100, 2))
        >>> state = kmeans(x, n_cluster=3, restarts=5, max_iters=50, key=key)
        >>> state.centers
        Array([[ 0.8450022 , -1.0791471 ],
                     [-0.7179966 ,  0.6372063 ],
                     [ 0.09818084, -0.25906876]], dtype=float32)
    """
    x = jnp.asarray(x, dtype=dtype)
    code_book, _, iters = kmeans_run(key, x, n_cluster, max_iters, restarts)
    assignment, _ = vq(x, code_book)
    totss = jnp.sum((x - jnp.mean(x, axis=0).reshape(1, -1)) ** 2)
    # Vector of within-cluster sum of squares, one component per cluster.
    withinss = jnp.array(  # pylint: disable=invalid-unary-operand-type
        [
            jnp.sum((x[assignment == i] - code_book[i].reshape(1, -1)) ** 2)
            for i in range(n_cluster)
        ]
    )
    # The between-cluster sum of squares,
    betwss = totss - jnp.sum(withinss)
    tot_withinss = jnp.sum(withinss)

    return KMeansState(
        n_clusters=n_cluster,
        centers=code_book,
        cluster=assignment,
        iters=iters,
        totss=totss,
        betwss=betwss,
        withinss=withinss,
        tot_withinss=tot_withinss,
    )


@filter_jit
def kmeans_run(key, points, k, max_iters, restarts):
    all_centroids, all_distortions, iters = vmap(
        lambda key: _kmeans_run(key, points, k, max_iters)
    )(jrandom.split(key, restarts))
    i = jnp.argmin(all_distortions)
    return all_centroids[i], all_distortions[i], iters[i]


@filter_jit
def _kmeans_run(key, points, k, max_iters, thresh=1e-5):
    def improve_centroids(val):
        prev_centroids, prev_distn, _, i = val
        assignment, distortions = vq(points, prev_centroids)

        counts = (
            (assignment[jnp.newaxis, :] == jnp.arange(k)[:, jnp.newaxis])
            .sum(axis=1, keepdims=True)
            .clip(min=1)
        )
        new_centroids = jnp.sum(
            jnp.where(
                assignment[:, jnp.newaxis, jnp.newaxis]
                == jnp.arange(k)[jnp.newaxis, :, jnp.newaxis],
                points[:, jnp.newaxis, :],
                0.0,
            ).astype(dtype=jnp.float32),
            axis=0,
        ) / jnp.asarray(counts, dtype=jnp.float32)

        return new_centroids, jnp.mean(distortions), prev_distn, i + 1

    initial_indices = jrandom.permutation(
        key, jnp.arange(points.shape[0]), independent=True
    )[:k]
    initial_val = improve_centroids((points[initial_indices, :], jnp.inf, None, 0))
    centroids, distortion, _, iters = lax.while_loop(
        lambda val: ((val[2] - val[1]) > thresh) & (val[3] < max_iters),
        improve_centroids,
        initial_val,
    )
    return centroids, distortion, iters



from functools import partial

import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import filter_jit
from jax import Array, jit

from ...distribution import pF
from ...hypothesis import HypoTest


class OneWayAnovaTest(HypoTest):
    df_between: int
    df_within: int
    df_tol: int
    ss_between: float
    ss_within: float
    ss_total: float
    ms_between: float
    ms_within: float

    def __init__(
        self,
        statistic=None,
        p_value=None,
        df_between=None,
        df_within=None,
        df_total=None,
        ss_between=None,
        ss_within=None,
        ss_total=None,
        ms_between=None,
        ms_within=None,
        name="One Way Anova Test",
    ):
        super().__init__(statistic=statistic, p_value=p_value, name=name)
        self.df_between = df_between
        self.df_within = df_within
        self.df_tol = df_total
        self.ss_between = ss_between
        self.ss_within = ss_within
        self.ss_total = ss_total
        self.ms_between = ms_between
        self.ms_within = ms_within

    def _summary(self):
        raise NotImplementedError


@partial(jit, static_argnames=("axis"))
def _square_of_sums(a, axis=0) -> Array:
    s = jnp.sum(a, axis, keepdims=True)
    return s * s


@partial(jit, static_argnames=("axis"))
def _sum_of_squares(a, axis=0) -> Array:
    return jnp.sum(a * a, axis=axis, keepdims=True)


def one_way(*samples, axis=0) -> OneWayAnovaTest:
    """Performs a one-way ANOVA test.

    Args:
        *samples: A sequence of samples to compare.
        axis (int): The axis along which the samples are compared.

    Returns:
        OneWayAnovaTest: The result of the one-way ANOVA test.

    Example:
        >>> import jax.numpy as jnp
        >>> from stamox.functions import one_way
        >>>  # 3 samples with 5 observations each
        >>> a = jnp.array([1, 2, 3, 4, 5])
        >>> b = jnp.array([2, 3, 4, 5, 6])
        >>> c = jnp.array([3, 4, 5, 6, 7])
        >>> one_way(a, b, c)

    """
    samples = [jnp.asarray(sample) for sample in samples]
    ngroups = len(samples)
    return _one_way(samples, ngroups, axis=axis)


@filter_jit
def _one_way(samples, ngroups, axis=0):
    alldata = jnp.concatenate(samples, axis=axis)
    N = alldata.shape[axis]
    offset = jnp.mean(alldata, axis=axis, keepdims=True)
    alldata -= offset
    normalized_ss = _square_of_sums(alldata, axis=axis) / N

    # total
    sstot = _sum_of_squares(alldata, axis=axis) - normalized_ss

    # between groups or namely treatments
    ssbn = jnp.sum(
        jnp.concatenate(
            jtu.tree_map(
                lambda s: _square_of_sums(s - offset, axis) / s.shape[0], samples
            ),
            axis,
        ),
        axis,
        keepdims=True,
    )
    ssbn -= normalized_ss

    # with in groups
    sswn = sstot - ssbn

    # degree of freedom
    dfbn = ngroups - 1
    dfwn = N - ngroups

    # mean squares
    msb = ssbn / dfbn
    msw = sswn / dfwn

    # F-ratio
    f = msb / msw
    f = f.squeeze(axis=axis)
    prob = 1 - pF(f, dfbn, dfwn)
    return OneWayAnovaTest(
        statistic=f,
        p_value=prob,
        df_between=dfbn,
        df_within=dfwn,
        df_total=N - 1,
        ss_between=ssbn[0],
        ss_within=sswn[0],
        ss_total=sstot[0],
        ms_between=msb[0],
        ms_within=msw[0],
    )



"""Test Basic Functions"""
import numpy as np

from stamox import Pipeable
from stamox.pipe_functions import mean, median, scale, sd, var


def test_mean():
    a = np.random.normal(size=(10000, 3))
    np.testing.assert_allclose(mean(a, axis=1), np.mean(a, axis=1))


def test_pipe_mean():
    a = np.random.normal(size=(10000, 3))
    p = Pipeable(a) >> mean(axis=1)
    np.testing.assert_allclose(p(a), np.mean(a, axis=1))


def test_median():
    a = np.random.normal(size=(10000, 3))
    np.testing.assert_allclose(median(a, axis=1), np.median(a, axis=1))


def test_pipe_median():
    a = np.random.normal(size=(10000, 3))
    p = Pipeable(a) >> median(axis=1)
    np.testing.assert_allclose(p(a), np.median(a, axis=1))


def test_std():
    a = np.random.normal(size=(10000, 3))
    np.testing.assert_allclose(sd(a, axis=1), np.std(a, axis=1))


def test_pipe_std():
    a = np.random.normal(size=(10000, 3))
    p = Pipeable(a) >> sd(axis=1)
    np.testing.assert_allclose(p(a), np.std(a, axis=1))


def test_var():
    a = np.random.normal(size=(10000, 3))
    np.testing.assert_allclose(var(a, axis=1), np.var(a, axis=1))


def test_pipe_var():
    a = np.random.normal(size=(10000, 3))
    p = Pipeable(a) >> var(axis=1)
    np.testing.assert_allclose(p(a), np.var(a, axis=1))


def test_scale():
    a = np.random.normal(size=(10000, 3)) * 2
    np.testing.assert_allclose(
        scale(a), (a - np.mean(a, axis=0)) / np.std(a, axis=0, ddof=1)
    )



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_jit
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Float, Int

from ._gamma import _dgamma, _pgamma, _qgamma
from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


def dchisq(
    x: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the chi-squared distribution.

    Args:
        x: A float or array-like object representing the values at which to evaluate the chi-squared distribution.
        df: The degrees of freedom for the chi-squared distribution.
        lower_tail: A boolean indicating whether to compute the lower tail of the chi-squared distribution (defaults to True).
        log_prob: A boolean indicating whether to return the log probability (defaults to False).
        dtype: The dtype of the output (defaults to float32).

    Returns:
        ArrayLike: The chi-squared distribution evaluated at `x`.

    Example:
        >>> dchisq(2.0, 3)
        Array(0.20755368, dtype=float32, weak_type=True)
    """
    x, dtype = _promote_dtype_to_floating(x, dtype)
    x = _check_clip_distribution_domain(x, 0.0, jnp.inf)
    grads = svmap_(_dgamma, x, df / 2, 1 / 2)
    grads = _post_process(grads, lower_tail, log_prob)
    return grads


def pchisq(
    q: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the chi-squared probability density function.

    Args:
        q (Union[float, array-like]): The value of the chi-squared variable.
        df (Union[int, float, array-like]): The degrees of freedom.
        lower_tail (bool): Whether to calculate the lower tail (default True).
        log_prob (bool): Whether to return the log probability (default False).
        dtype (dtype): The dtype of the output (default jnp.float_).

    Returns:
        ArrayLike: The chi-squared probability density function.

    Example:
        >>> pchisq(2.0, 3)
        Array(0.42759317, dtype=float32, weak_type=True)
    """
    q = jnp.asarray(q, dtype=dtype)
    q = _check_clip_distribution_domain(q, 0.0, jnp.inf)
    p = svmap_(_pgamma, q, df / 2, 1 / 2)
    p = _post_process(p, lower_tail, log_prob)
    return p


def qchisq(
    p: Union[Float, ArrayLike],
    df: Union[Int, Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the inverse of the chi-squared cumulative distribution function.

    Args:
        p (Union[Float, ArrayLike]): Probability value or array of probability values.
        df (Union[Int, Float, ArrayLike]): Degrees of freedom.
        lower_tail (bool, optional): If True (default), probabilities are P[X ≤ x], otherwise, P[X > x].
        log_prob (bool, optional): If True, probabilities are given as log(p).
        dtype (dtype, optional): The dtype of the output (default jnp.float_).

    Returns:
        ArrayLike: The quantiles corresponding to the given probabilities.

    Example:
        >>> qchisq(0.95, 10)
        Array(18.307034, dtype=float32)
    """
    p = jnp.asarray(p, dtype=dtype)
    p = _check_clip_probability(p, lower_tail, log_prob)
    q = svmap_(_qgamma, p, df / 2, 1 / 2)
    return q


@filter_jit
def _rchisq(
    key: KeyArray,
    df: Union[Int, Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    if sample_shape is None:
        sample_shape = jnp.shape(df)
    df = jnp.broadcast_to(df, sample_shape)
    return jrand.chisquare(key, df, shape=sample_shape, dtype=dtype)


def rchisq(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    df: Union[Int, Float, ArrayLike] = None,
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """
    Generates random variates from the chi-squared distribution.

    Args:
        key (KeyArray): Random key to generate the random numbers.
        sample_shape (Optional[Shape], optional): Shape of the output array. Defaults to None.
        df (Union[Int, Float, ArrayLike], optional): Degrees of freedom. Defaults to None.
        lower_tail (bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (dtype, optional): The dtype of the output (default float_).

    Returns:
        ArrayLike: Random variates from the chi-squared distribution.

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> rchisq(key, df=2)
        Array(1.982825, dtype=float32)
    """
    rvs = _rchisq(key, df, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit
from jax import lax
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Float

from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _pweibull(
    x: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
):
    concentration = lax.convert_element_type(concentration, x.dtype)
    scale = lax.convert_element_type(scale, x.dtype)
    scaled_x = lax.div(x, scale)
    powered = jnp.float_power(scaled_x, concentration)
    return 1 - jnp.exp(-powered)


def pweibull(
    q: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the cumulative distribution function of the Weibull distribution.

    Args:
        q (Union[Float, ArrayLike]): The value at which to evaluate the CDF.
        concentration (Union[Float, ArrayLike], optional): The concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.

    Returns:
        Array: The cumulative distribution function of the Weibull distribution evaluated at `q`.

    Example:
        >>> pweibull(1.0, concentration=1.0, scale=1.0)
    """
    q, _ = _promote_dtype_to_floating(q, dtype)
    q = _check_clip_distribution_domain(q, lower=0.0)
    p = svmap_(_pweibull, q, concentration, scale)
    p = _post_process(p, lower_tail=lower_tail, log_prob=log_prob)
    return p


_dweibull = filter_grad(filter_jit(_pweibull))


def dweibull(
    x, concentration=0.0, scale=1.0, lower_tail=True, log_prob=False, dtype=None
) -> ArrayLike:
    """Computes the probability density function of the Weibull distribution.

    Args:
        x (Union[Float, ArrayLike]): The value at which to evaluate the PDF.
        concentration (Union[Float, ArrayLike], optional): The concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the CDF. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (Optional[jnp.dtype], optional): The dtype of the output. Defaults to None.


    Returns:
        Array: The probability density function of the Weibull distribution evaluated at `x`.

    Example:
        >>> dweibull(0.5, 1.0, 1.0)
    """
    x, _ = _promote_dtype_to_floating(x, dtype)
    x = _check_clip_distribution_domain(x, lower=0.0)
    grads = svmap_(_dweibull, x, concentration, scale)
    grads = _post_process(grads, lower_tail=lower_tail, log_prob=log_prob)
    return grads


@filter_jit
def _qweibull(
    q: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
) -> ArrayLike:
    concentration = lax.convert_element_type(concentration, q.dtype)
    scale = lax.convert_element_type(scale, q.dtype)
    one = lax.convert_element_type(1, q.dtype)
    nlog_q = -lax.log(lax.sub(one, q))
    inv_concentration = lax.div(one, concentration)
    powerd = jnp.float_power(nlog_q, inv_concentration)
    x = lax.mul(powerd, scale)
    return x


def qweibull(
    p: Union[Float, ArrayLike],
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the quantile function of the Weibull distribution.

    Args:
        p (Union[Float, ArrayLike]): The quantiles to compute.
        concentration (Union[Float, ArrayLike], optional): The concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): The scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to compute the lower tail of the distribution. Defaults to True.
        log_prob (bool, optional): Whether to compute the log probability of the distribution. Defaults to False.
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to jnp.float_.


    Returns:
        Array: The computed quantiles.

    Example:
        >>> qweibull(0.5, 1.0, 1.0)
    """
    p, _ = _promote_dtype_to_floating(p, dtype)
    p = _check_clip_probability(p, lower_tail, log_prob)
    q = svmap_(_qweibull, p, concentration, scale)
    return q


@filter_jit
def _rweibull(
    key: KeyArray,
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
):
    return jrand.weibull_min(key, scale, concentration, sample_shape, dtype=dtype)


def rweibull(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    concentration: Union[Float, ArrayLike] = 0.0,
    scale: Union[Float, ArrayLike] = 1.0,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generates samples from the Weibull distribution.

    Args:
        key (KeyArray): Random key used for generating random numbers.
        sample_shape (Optional[Shape], optional): Shape of the output sample. Defaults to None.
        concentration (Union[Float, ArrayLike], optional): Concentration parameter of the Weibull distribution. Defaults to 0.0.
        scale (Union[Float, ArrayLike], optional): Scale parameter of the Weibull distribution. Defaults to 1.0.
        lower_tail (bool, optional): Whether to return the lower tail probability. Defaults to True.
        log_prob (bool, optional): Whether to return the log probability. Defaults to False.
        dtype (Optional[jnp.dtype], optional): The dtype of the output. Defaults to jnp.float_.


    Returns:
        rvs (ArrayLike): Probability of the Weibull distribution.

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> rweibull(key, 1.0, 1.0)
    """
    rvs = _rweibull(key, concentration, scale, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail=lower_tail, log_prob=log_prob)
    return rvs



from stamox.sample._bootstrap import bootstrap, bootstrap_sample
from stamox.sample._jackknife import jackknife, jackknife_sample


__all__ = [
    "bootstrap",
    "bootstrap_sample",
    "jackknife",
    "jackknife_sample",
]



from typing import Callable, Optional, TypeVar

import equinox as eqx


T = TypeVar("T")


class Functional(eqx.Module):
    """General Function"""

    _name: str
    _fn: Callable[..., T]
    _pipe_type: str

    def __init__(
        self,
        fn: Optional[Callable[..., T]] = None,
        pipe_type: str = "normal",
        name: str = "Func",
    ):
        """Make a General Function.

        Args:
            name (str, optional): Name of the function. Defaults to "Func".
            pipe_type: Type of the function.(vmap or pmap or jit or normal) Defaults to "normal".
            fn (Optional[Callable|None], optional): Callable object.
        """
        super().__init__()
        self._name = name
        self._fn = fn
        self._pipe_type = pipe_type

    @property
    def name(self):
        """Get the name of the function."""
        return self._name

    @property
    def func(self):
        """Get the function."""
        return self._fn

    @property
    def piep_type(self):
        """Get the type of the function."""
        return self._pipe_type

    def desc(self):
        """Description for the function."""
        return self.__repr__()

    def __call__(self, *args, **kwargs):
        """Call the function with given arguments."""
        if self._fn is None:
            raise ValueError("No Callable Function to Call")
        return self._fn(*args, **kwargs)

    def __rshift__(self, _next: Callable):
        """Make Pipe.

        Create a pipe between this function and the next one.

        Args:
            _next (Functional): The next function in the pipe.

        Returns:
            Pipe: A pipe between this function and the next one.
        """
        from .pipe import Pipe

        if not isinstance(_next, Functional):
            if hasattr(_next, "__name__"):
                _next = Functional(name=_next.__name__, fn=_next)
            else:
                _next = Functional(name="Function", fn=_next)

        return Pipe([self, _next])


class StateFunc(Functional):
    """Class for state function.

    Args:
        name (str): Name of the state function.
        fn (Optional[Callable]): Function to be called.
    """

    def __init__(
        self,
        fn: Optional[Callable[..., T]] = None,
        name: str = "State",
    ):
        """Initialize the state function."""
        super().__init__(fn=fn, name=name)

    def __repr__(self):
        """Return a string representation of the state function."""
        return super().__repr__()

    def _tree_flatten(self):
        """Flatten the tree structure of the state function."""
        return super()._tree_flatten()

    def _summary(self):
        """Print a summary of the state function."""
        pass

    def _predict(self, *args, **kwargs):
        """Predict"""
        pass

    def __call__(self, *args, **kwargs):
        """Call the state function."""
        return self.func(*args, **kwargs)



from typing import Optional

import jax.numpy as jnp
from equinox import filter_jit
from jaxtyping import ArrayLike


def pearsonr(x: ArrayLike, y: Optional[ArrayLike] = None, axis: int = 0) -> ArrayLike:
    """Computes Pearson correlation coefficient for two arrays.

    Args:
        x: An array-like object containing the first set of data.
        y: An optional array-like object containing the second set of data. If not
        provided, `x` is assumed to contain two sets of data.
        axis: The axis along which the correlation coefficient should be computed.
        Must be 0 or 1. Defaults to 0.

    Returns:
        An array-like object containing the Pearson correlation coefficient.

    Examples:
        >>> import jax.numpy as jnp
        >>> from stamox.functions import pearsonr
        >>> x = jnp.array([1, 2, 3, 4, 5])
        >>> y = jnp.array([5, 6, 7, 8, 7])
        >>> pearsonr(x, y)
        Array(0.8320503, dtype=float32)
    """
    if axis is not None and axis > 1:
        raise ValueError(
            "pearsonr only handles 1-D or 2-D arrays, "
            "supplied axis argument {}, please use only "
            "values 0, 1 or None for axis".format(axis)
        )
    x = jnp.asarray(x)
    if axis is None:
        x = x.ravel()
        axis_out = 0
    else:
        axis_out = axis

    if x.ndim > 2:
        raise ValueError("pearsonr only handles 1-D or 2-D arrays")

    if y is None:
        if x.ndim < 2:
            raise ValueError("`pearsonr` needs at least 2 " "variables to compare")
    else:
        y = jnp.asarray(y, dtype=x.dtype)
        if axis is None:
            y = y.ravel()
        if axis_out == 0:
            x = jnp.column_stack((x, y))
        else:
            x = jnp.row_stack((x, y))

    return _pearsonr(x, axis)


@filter_jit
def _pearsonr(x, axis):
    coef = jnp.corrcoef(x, rowvar=axis)
    if coef.shape == (2, 2):
        coef = coef[1, 0]
    return coef



from typing import Optional, Union

import jax.numpy as jnp
import jax.random as jrand
from equinox import filter_grad, filter_jit
from jax import lax
from jax._src.random import KeyArray, Shape
from jaxtyping import ArrayLike, Float
from tensorflow_probability.substrates.jax.math import special as tfp_special

from ._utils import (
    _check_clip_distribution_domain,
    _check_clip_probability,
    _post_process,
    _promote_dtype_to_floating,
    svmap_,
)


@filter_jit
def _pbeta(
    x: Union[Float, ArrayLike], a: Union[Float, ArrayLike], b: Union[Float, ArrayLike]
):
    dtype = lax.dtype(x)
    a = jnp.asarray(a, dtype=dtype)
    b = jnp.asarray(b, dtype=dtype)
    return tfp_special.betainc(a, b, x)


def pbeta(
    q: Union[Float, ArrayLike],
    a: Union[Float, ArrayLike],
    b: Union[Float, ArrayLike],
    lower_tail=True,
    log_prob=False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the cumulative distribution function of the beta distribution.

    Args:
        q (Union[Float, ArrayLike]): Quantiles.
        a (Union[Float, ArrayLike]): Shape parameter.
        b (Union[Float, ArrayLike]): Shape parameter.
        lower_tail (bool, optional): If True (default), probabilities are P[X ≤ x], otherwise, P[X > x].
        log_prob (bool, optional): If True, probabilities are given as log(P).
        dtype (jnp.dtype, optional): The dtype of the output. Defaults to None.

    Returns:
        ArrayLike: The probability or log of the probability for each quantile.

    Example:
        >>> q = jnp.array([0.1, 0.5, 0.9])
        >>> a = 2.0
        >>> b = 3.0
        >>> pbeta(q, a, b)
        Array([0.05230004, 0.68749976, 0.9963    ], dtype=float32)
    """
    q, dtype = _promote_dtype_to_floating(q, dtype)
    q = _check_clip_distribution_domain(q, 0.0, 1.0)
    p = svmap_(_pbeta, q, a, b)
    p = _post_process(p, lower_tail, log_prob)
    return p


_dbeta = filter_jit(filter_grad(_pbeta))


def dbeta(
    x: Union[Float, ArrayLike],
    a: Union[Float, ArrayLike],
    b: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Calculates the probability density function of the beta distribution.

    Args:
      x: A float or array-like object representing the value(s) at which to evaluate the PDF.
      a: A float or array-like object representing the shape parameter of the beta distribution.
      b: A float or array-like object representing the scale parameter of the beta distribution.
      lower_tail: A boolean indicating whether to calculate the lower tail (default True).
      log_prob: A boolean indicating whether to return the logarithm of the PDF (default False).
      dtype: The dtype of the output. Defaults to None.

    Returns:
      ArrayLike: The probability density function of the beta distribution evaluated at x.

    Example:
        >>> dbeta(0.5, 2, 3, lower_tail=True, log_prob=False)
        Array(1.4999996, dtype=float32, weak_type=True)
    """
    x, dtype = _promote_dtype_to_floating(x, dtype)
    x = _check_clip_distribution_domain(x, 0.0, 1.0)
    d = svmap_(_dbeta, x, a, b)
    d = _post_process(d, lower_tail, log_prob)
    return d


@filter_jit
def _qbeta(
    p: Union[Float, ArrayLike], a: Union[Float, ArrayLike], b: Union[Float, ArrayLike]
):
    return tfp_special.betaincinv(a, b, p)


def qbeta(
    p: Union[Float, ArrayLike],
    a: Union[Float, ArrayLike],
    b: Union[Float, ArrayLike],
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Computes the quantile of beta distribution function.

    Args:
        p: A float or array-like object representing the quantile.
        a: A float or array-like object representing the alpha parameter.
        b: A float or array-like object representing the beta parameter.
        lower_tail: A boolean indicating whether to compute the lower tail of the
        distribution (defaults to True).
        log_prob: A boolean indicating whether to compute the log probability
        (defaults to False).
        dtype: The dtype of the output. Defaults to None.

    Returns:
        ArrayLike: The value of the beta distribution at the given quantile.

    Example:
        >>> qbeta(0.5, 2, 3)
        Array(0.38572744, dtype=float32)
    """
    p, dtype = _promote_dtype_to_floating(p, dtype)
    p = _check_clip_probability(p, lower_tail=lower_tail, log_prob=log_prob)
    x = svmap_(_qbeta, p, a, b)
    return x


def rbeta(
    key: KeyArray,
    sample_shape: Optional[Shape] = None,
    a: Union[Float, ArrayLike] = None,
    b: Union[Float, ArrayLike] = None,
    lower_tail: bool = True,
    log_prob: bool = False,
    dtype=jnp.float_,
) -> ArrayLike:
    """Generates random numbers from the Beta distribution.

    Args:
        key: A PRNGKey used for random number generation.
        sample_shape: An optional shape for the output samples.
        a: The shape parameter of the Beta distribution. Can be either a float or an array-like object.
        b: The scale parameter of the Beta distribution. Can be either a float or an array-like object.
        lower_tail: Whether to return the lower tail probability (defaults to True).
        log_prob: Whether to return the log probability (defaults to False).
        dtype: The dtype of the output. Defaults to jnp.float32.

    Returns:
        ArrayLike: Random numbers from the Beta distribution.

    Example:
        >>> key = jax.random.PRNGKey(0)
        >>> rbeta(key, sample_shape=(3,), a=2, b=3)
        Array([0.02809353, 0.13760717, 0.49360353], dtype=float32)
    """
    rvs = _rbeta(key, a, b, sample_shape, dtype=dtype)
    rvs = _post_process(rvs, lower_tail, log_prob)
    return rvs


@filter_jit
def _rbeta(
    key: KeyArray,
    a: Union[Float, ArrayLike],
    b: Union[Float, ArrayLike],
    sample_shape: Optional[Shape] = None,
    dtype=jnp.float_,
) -> ArrayLike:
    if sample_shape is None:
        sample_shape = jnp.broadcast_shapes(jnp.shape(a), jnp.shape(b))
    a = jnp.broadcast_to(a, sample_shape)
    b = jnp.broadcast_to(b, sample_shape)
    return jrand.beta(key, a, b, sample_shape, dtype=dtype)



import jax.numpy as jnp
from jaxtyping import ArrayLike

from ..core import StateFunc


class RegState(StateFunc):
    in_features: int
    out_features: int
    _coefs: ArrayLike
    _df_resid: int
    _df_model: int
    _n_obs: int
    _resid: ArrayLike
    _fitted_values: ArrayLike
    _rank: int
    _dtype: jnp.dtype

    def __init__(
        self,
        in_features,
        out_features,
        coefs=None,
        df_resid=None,
        df_model=None,
        n_obs=None,
        resid=None,
        fitted_values=None,
        rank=None,
        dtype=jnp.float32,
        name="RegState",
    ):
        super().__init__(name=name, fn=None)
        self.in_features = in_features
        self.out_features = out_features
        if coefs is None:
            coefs = jnp.zeros((in_features, out_features))
        self._coefs = coefs
        self._df_resid = df_resid
        self._df_model = df_model
        self._n_obs = n_obs
        self._resid = resid
        self._fitted_values = fitted_values
        self._rank = rank
        self._dtype = dtype

    @property
    def params(self):
        return self._coefs

    @property
    def coefs(self):
        return self._coefs.ravel()

    @property
    def coefs_X(self):
        return self.params[1:, :]

    @property
    def intercept(self):
        return self.params[0, :].reshape(-1, 1)

    @property
    def df_resid(self):
        return self._df_resid

    @property
    def df_model(self):
        return self._df_model

    @property
    def n_obs(self):
        return self._n_obs

    @property
    def resid(self):
        return self._resid

    @property
    def fitted_values(self):
        return self._fitted_values

    @property
    def rank(self):
        return self._rank

    @property
    def dtype(self):
        return self._dtype
    
    def _transform(self, X):
        if X.shape[1] == self.in_features:
            return jnp.matmul(X, self.params)
        else:
            return jnp.matmul(X, self.coefs_X) + self.intercept

    def _predict(self, X):
        return self._transform(X)

    def _summary(self):
        return "Not Implement Yet"





from stamox.hypothesis._bartlett import bartlett_test, BartlettTest
from stamox.hypothesis._base import HypoTest
from stamox.hypothesis._durbin_watson import durbin_watson_test, DurbinWatsonTest
from stamox.hypothesis._friedman import friedman_test, FriedmanTest


__all__ = [
    "HypoTest",
    "bartlett_test",
    "BartlettTest",
    "durbin_watson_test",
    "DurbinWatsonTest",
    "friedman_test",
    "FriedmanTest",
]



import jax.numpy as jnp
from equinox import filter_jit
from jax import lax, vmap
from jaxtyping import ArrayLike

from ..core import StateFunc


class PCAState(StateFunc):
    n_components: int
    components: ArrayLike
    eigen_vals: ArrayLike
    mean: ArrayLike

    def __init__(self, n_components, components, eigen_vals, mean):
        super().__init__(name="PCAState", fn=None)
        self.n_components = n_components
        self.components = components
        self.mean = mean
        self.eigen_vals = eigen_vals


def princomp(x: ArrayLike, n_components: int) -> PCAState:
    """Performs principal component analysis (PCA) on the given array.

    Args:
        x (ArrayLike): The input array of shape (n_samples, n_features).
        n_components (int): Number of components to keep.

    Returns:
        PCAState: A namedtuple containing the results of the PCA.

    Example:
        >>> import jax.numpy as jnp
        >>> from stamox.functions import princomp
        >>> X = jnp.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]], dtype=jnp.float32)
        >>> pca_state = princomp(X, n_components=2)
        >>> pca_state.components
        Array([[-0.8384922,  0.5449136],
            [-0.5449136, -0.8384922]], dtype=float32)
    """
    dtype = lax.dtype(x)
    x = jnp.asarray(x, dtype=dtype)
    n_components = int(n_components)
    if n_components < 1:
        raise ValueError(
            f"n_components={n_components} must be greater than or equal to 1"
        )
    if n_components > x.shape[1]:
        raise ValueError(
            f"n_components={n_components} must be less than or equal to the number of features={x.shape[1]}"
        )
    return _pca(x, n_components)


@filter_jit
def _pca(X, n_components):
    # Subtract the mean of each feature column to center the data around the origin
    _mean = jnp.mean(X, axis=0)
    X_meaned = vmap(lambda x: x - _mean)(X)
    # Computing the covariance matrix
    cov_mat = jnp.cov(X_meaned.T)

    # Computing the eigenvalues and eigenvectors of the covariance matrix
    eigen_vals, eigen_vecs = jnp.linalg.eigh(cov_mat)

    # Sorting the eigenvalues in descending order and selecting the top n_components eigenvectors
    sorted_idx = eigen_vals.argsort()[::-1]
    sorted_eigen_vals = eigen_vals[sorted_idx]
    sorted_eigen_vecs = eigen_vecs[:, sorted_idx]
    eigenvectors_subset = sorted_eigen_vecs[:, :n_components]

    return PCAState(
        n_components=n_components,
        components=eigenvectors_subset,
        mean=_mean,
        eigen_vals=sorted_eigen_vals,
    )
